"""Fast Fourier transforms (:mod:`fluiddyn.calcul.easypyfft`)
=============================================================

.. autofunction:: fftw_grid_size

Provides classes for performing fft in 1, 2, and 3 dimensions:

.. autoclass:: FFTP2D
   :members:

.. autoclass:: FFTW2DReal2Complex
   :members:

.. autoclass:: FFTW3DReal2Complex
   :members:

.. autoclass:: FFTW1D
   :members:

.. autoclass:: FFTW1DReal2Complex
   :members:

"""

import os
import numpy as np
from copy import copy
from time import time

try:
    import scipy.fftpack as fftp
except ImportError:
    pass

if "OMP_NUM_THREADS" in os.environ:
    nthreads = int(os.environ["OMP_NUM_THREADS"])
else:
    nthreads = 1

from ..util.mpi import printby0


def fftw_grid_size(nk, bases=[2, 3, 5, 7, 11, 13], debug=False):
    """Find the closest multiple of prime powers greater than or equal to nk
    using Mixed Integer Linear Programming (MILP). Useful while setting the
    grid-size to be compatible with FFTW.

    Parameters
    ----------
    nk : int
        Lower bound for the spectral grid size.

    bases : array-like, optional
        List of bases, typically prime numbers.

    debug : bool, optional
        Print useful messages.

    Returns
    -------
    int

    """
    if {2, 3, 5} == set(bases):
        if debug:
            print("Using scipy.fftpack.next_fast_len")
        return fftp.next_fast_len(nk)

    elif {2, 3, 5, 7, 11, 13} == set(bases):
        try:
            import pyfftw

            return pyfftw.next_fast_len(nk)

        except (ImportError, AttributeError):
            pass
        else:
            if debug:
                print("Using pyfftw.next_fast_len")

    if not {2, 3, 5, 7, 11, 13}.issuperset(bases):
        raise ValueError(
            "FFTW only supports bases which are a subset of "
            "{2, 3, 5, 7, 11, 13}."
        )

    import pulp

    prob = pulp.LpProblem("FFTW Grid-size Problem")

    bases = np.array(bases)
    bases_order1 = bases[bases < 10]
    bases_order2 = bases[bases >= 10]

    exp_max = np.ceil(np.log2(nk))
    exps = pulp.LpVariable.dicts(
        "exponent_o1", bases_order1, 0, exp_max, cat=pulp.LpInteger
    )
    exps.update(
        pulp.LpVariable.dicts(
            "exponent_o2", bases_order2, 0, 1, cat=pulp.LpInteger
        )
    )

    log_nk_new = pulp.LpVariable("log_grid_size", 0)
    log_nk_new = pulp.lpDot(exps.values(), np.log(bases))

    prob += log_nk_new  # Target to be minimized
    # Subject to:
    prob += log_nk_new >= np.log(nk), "T1"
    if {11, 13}.issubset(bases):
        prob += exps[11] + exps[13] <= 1, "T2"

    if debug:
        print("bases =", bases)
        print("exponents =", exps)
        print("log_nk_new =", log_nk_new)
    # prob.writeLP("FFTWGridSizeOptimizationModel.lp")

    prob.solve()

    if debug:
        print("Status:", pulp.LpStatus[prob.status])
        for v in prob.variables():
            print(v.name, "=", v.varValue)

    if pulp.LpStatus[prob.status] == "Infeasible":
        raise ValueError(f"Not enough bases: {bases}")

    exps_solution = [v.varValue for v in prob.variables()]
    nk_new = np.prod(np.power(bases, exps_solution))
    return int(nk_new)


class BaseFFT:
    def run_tests(self):
        arr = np.random.rand(*self.shapeX)
        arr_fft = self.fft(arr)
        arr = self.ifft(arr_fft)
        arr_fft = self.fft(arr)

        nrj = self.compute_energy_from_spatial(arr)
        nrj_fft = self.compute_energy_from_Fourier(arr_fft)

        assert np.allclose(nrj, nrj_fft)

        arr2_fft = np.zeros(self.shapeK, dtype=np.complex128)
        self.fft_as_arg(arr, arr2_fft)
        nrj2_fft = self.compute_energy_from_Fourier(arr2_fft)
        assert np.allclose(nrj, nrj2_fft)

        arr2 = np.empty(self.shapeX)
        self.ifft_as_arg(arr_fft, arr2)
        nrj2 = self.compute_energy_from_spatial(arr2)
        assert np.allclose(nrj, nrj2)

    def run_benchs(self, nb_time_execute=10):

        arr = np.zeros(self.shapeX)
        arr_fft = np.zeros(self.shapeK, dtype=np.complex128)

        times = []
        for i in range(nb_time_execute):
            t_start = time()
            self.fft_as_arg(arr, arr_fft)
            times.append(time() - t_start)

        time_fft = np.mean(times)

        times = []
        for i in range(nb_time_execute):
            t_start = time()
            self.ifft_as_arg(arr_fft, arr)
            times.append(time() - t_start)

        time_ifft = np.mean(times)

        name = self.__class__.__name__
        printby0(
            "Internal bench (" + name + ")\n"
            "time fft ({}):  {:.6f} s\n".format(name, time_fft)
            + f"time ifft ({name}): {time_ifft:.6f} s"
        )

        return time_fft, time_ifft

    def get_short_name(self):
        return self.__class__.__name__.lower()

    def compute_energy_from_X(self, fieldX):
        return np.mean(fieldX ** 2 / 2.0)

    def get_local_size_X(self):
        return np.prod(self.shapeX)

    def get_shapeK_seq(self):
        return self.shapeK

    get_shapeK_loc = get_shapeK_seq

    def get_shapeX_seq(self):
        return self.shapeX

    get_shapeX_loc = get_shapeX_seq


class FFTP2D(BaseFFT):
    """ A class to use fftp """

    def __init__(self, nx, ny):
        if nx % 2 != 0 or ny % 2 != 0:
            raise ValueError("nx and ny should be even")

        self.nx = nx
        self.ny = ny
        self.shapeX = (ny, nx)
        self.nkx = int(float(nx) / 2 + 1)
        self.shapeK = self.shapeK_seq = self.shapeK_loc = (ny, self.nkx)
        self.coef_norm = nx * ny

        self.fft2d = self.fft
        self.ifft2d = self.ifft

    def fft(self, ff):
        if not (isinstance(ff[0, 0], float)):
            print("Warning: not array of floats")
        big_ff_fft = fftp.fft2(ff) / self.coef_norm
        small_ff_fft = big_ff_fft[:, 0 : self.nkx]
        return small_ff_fft

    def ifft(self, small_ff_fft, ARG_IS_COMPLEX=False):
        if not (isinstance(small_ff_fft[0, 0], complex)):
            print("Warning: not array of complexes")
        # print('small_ff_fft\n', small_ff_fft)
        big_ff_fft = np.empty(self.shapeX, dtype=np.complex128)
        big_ff_fft[:, 0 : self.nkx] = small_ff_fft
        for iky in range(self.ny):
            big_ff_fft[iky, self.nkx :] = small_ff_fft[
                -iky, self.nkx - 2 : 0 : -1
            ].conj()

        # print('big_ff_fft final\n', big_ff_fft)
        result_ifft = fftp.ifft2(big_ff_fft * self.coef_norm)
        if np.max(np.imag(result_ifft)) > 10 ** (-8):
            print(
                "ifft2: imaginary part of ifft not equal to zero,",
                np.max(np.imag(result_ifft)),
            )
        return np.real(result_ifft)

    def fft_as_arg(self, field, field_fft):
        field_fft[:] = self.fft(field)

    def ifft_as_arg(self, field_fft, field):
        field[:] = self.ifft(field_fft)

    def compute_energy_from_Fourier(self, ff_fft):
        return (
            np.sum(abs(ff_fft[:, 0]) ** 2 + abs(ff_fft[:, -1]) ** 2)
            + 2 * np.sum(abs(ff_fft[:, 1:-1]) ** 2)
        ) / 2

    def compute_energy_from_spatial(self, ff):
        return np.mean(abs(ff) ** 2) / 2


class BasePyFFT(BaseFFT):
    def __init__(self, shapeX):
        try:
            import pyfftw
        except ImportError as err:
            raise ImportError(
                "ImportError {0}. Instead fftpack can be used (?)", err
            )

        shapeK = list(shapeX)
        shapeK[-1] = shapeK[-1] // 2 + 1
        shapeK = tuple(shapeK)

        self.shapeX = shapeX
        self.shapeK = self.shapeK_seq = self.shapeK_loc = shapeK

        self.empty_aligned = pyfftw.empty_aligned
        self.arrayX = pyfftw.empty_aligned(shapeX, np.float64)
        self.arrayK = pyfftw.empty_aligned(shapeK, np.complex128)

        axes = tuple(range(len(shapeX)))

        self.fftplan = pyfftw.FFTW(
            input_array=self.arrayX,
            output_array=self.arrayK,
            axes=axes,
            direction="FFTW_FORWARD",
            threads=nthreads,
        )
        self.ifftplan = pyfftw.FFTW(
            input_array=self.arrayK,
            output_array=self.arrayX,
            axes=axes,
            direction="FFTW_BACKWARD",
            threads=nthreads,
        )

        self.coef_norm = np.prod(shapeX)
        self.inv_coef_norm = 1.0 / self.coef_norm

    def fft(self, fieldX):
        fieldK = self.empty_aligned(self.shapeK, np.complex128)
        self.fftplan(
            input_array=fieldX, output_array=fieldK, normalise_idft=False
        )
        return fieldK / self.coef_norm

    def ifft(self, fieldK):
        fieldX = self.empty_aligned(self.shapeX, np.float64)
        # This copy is needed because FFTW_DESTROY_INPUT is used.
        # See pyfftw.readthedocs.io/en/latest/source/pyfftw/pyfftw.html
        self.arrayK[:] = fieldK
        self.ifftplan(
            input_array=self.arrayK, output_array=fieldX, normalise_idft=False
        )
        return fieldX

    def fft_as_arg(self, fieldX, fieldK):
        self.fftplan(
            input_array=fieldX, output_array=fieldK, normalise_idft=False
        )
        fieldK *= self.inv_coef_norm

    def ifft_as_arg(self, fieldK, fieldX):
        # This copy is needed because FFTW_DESTROY_INPUT is used.
        # See pyfftw.readthedocs.io/en/latest/source/pyfftw/pyfftw.html
        # fieldK = fieldK.copy()
        # self.ifftplan(input_array=fieldK, output_array=fieldX,
        # this seems faster (but it could depend on the size)
        self.arrayK[:] = fieldK
        self.ifftplan(
            input_array=self.arrayK, output_array=fieldX, normalise_idft=False
        )

    def ifft_as_arg_destroy(self, fieldK, fieldX):
        self.ifftplan(
            input_array=fieldK, output_array=fieldX, normalise_idft=False
        )

    def compute_energy_from_Fourier(self, ff_fft):
        result = self.sum_wavenumbers(abs(ff_fft) ** 2) / 2
        return result

    compute_energy_from_K = compute_energy_from_Fourier

    def compute_energy_from_spatial(self, ff):
        return np.mean(abs(ff) ** 2) / 2

    def project_fft_on_realX(self, ff_fft):
        return self.fft(self.ifft(ff_fft))

    def get_is_transposed(self):
        return False

    def create_arrayX(self, value=None):
        """Return a constant array in real space."""
        shapeX = self.shapeX
        field = self.empty_aligned(shapeX)
        if value is not None:
            field.fill(value)
        return field

    def create_arrayK(self, value=None):
        """Return a constant array in real space."""
        shapeK = self.shapeK
        field = self.empty_aligned(shapeK, dtype=np.complex128)
        if value is not None:
            field.fill(value)
        return field


class FFTW2DReal2Complex(BasePyFFT):
    """ A class to use fftw """

    def __init__(self, nx, ny):
        shapeX = (ny, nx)
        super().__init__(shapeX)
        self.fft2d = self.fft
        self.ifft2d = self.ifft

    def sum_wavenumbers(self, ff_fft):
        if self.shapeX[1] % 2 == 0:
            return (
                np.sum(ff_fft[:, 0])
                + np.sum(ff_fft[:, -1])
                + 2 * np.sum(ff_fft[:, 1:-1])
            )

        else:
            return np.sum(ff_fft[:, 0]) + 2 * np.sum(ff_fft[:, 1:])

    def get_seq_indices_first_K(self):
        return 0, 0

    def get_seq_indices_first_X(self):
        return 0, 0

    def get_x_adim_loc(self):
        """Get the coordinates of the points stored locally.

        Returns
        -------

        x0loc : np.ndarray

        x1loc : np.ndarray

        The indices correspond to the index of the dimension in real space.
        """
        nyseq, nxseq = self.get_shapeX_seq()

        ix0_start, ix1_start = self.get_seq_indices_first_X()
        nx0loc, nx1loc = self.get_shapeX_loc()

        x0loc = np.array(range(ix0_start, ix0_start + nx0loc))
        x1loc = np.array(range(ix1_start, ix1_start + nx1loc))

        return x0loc, x1loc

    def get_k_adim_loc(self):
        """Get the non-dimensional wavenumbers stored locally.

        Returns
        -------

        k0_adim_loc : np.ndarray

        k1_adim_loc : np.ndarray

        The indices correspond to the index of the dimension in spectral space.
        """

        nyseq, nxseq = self.get_shapeX_seq()

        kyseq = np.array(
            list(range(nyseq // 2 + 1)) + list(range(-nyseq // 2 + 1, 0))
        )
        kxseq = np.array(range(nxseq // 2 + 1))

        if self.get_is_transposed():
            k0seq, k1seq = kxseq, kyseq
        else:
            k0seq, k1seq = kyseq, kxseq

        ik0_start, ik1_start = self.get_seq_indices_first_K()
        nk0loc, nk1loc = self.get_shapeK_loc()

        k0_adim_loc = k0seq[ik0_start : ik0_start + nk0loc]
        k1_adim_loc = k1seq[ik1_start : ik1_start + nk1loc]

        return k0_adim_loc, k1_adim_loc


class FFTW3DReal2Complex(BasePyFFT):
    """ A class to use fftw """

    def __init__(self, nx, ny, nz):
        shapeX = (nz, ny, nx)
        super().__init__(shapeX)
        self.fft3d = self.fft
        self.ifft3d = self.ifft

    def sum_wavenumbers(self, ff_fft):
        if self.shapeX[2] % 2 == 0:
            return (
                np.sum(ff_fft[:, :, 0])
                + np.sum(ff_fft[:, :, -1])
                + 2 * np.sum(ff_fft[:, :, 1:-1])
            )

        else:
            return np.sum(ff_fft[:, :, 0]) + 2 * np.sum(ff_fft[:, :, 1:])

    def get_k_adim(self):
        nK0, nK1, nK2 = self.shapeK
        kz_adim_max = nK0 // 2
        kz_adim_min = -((nK0 - 1) // 2)
        ky_adim_max = nK1 // 2
        ky_adim_min = -((nK1 - 1) // 2)
        return (
            np.r_[0 : kz_adim_max + 1, kz_adim_min:0],
            np.r_[0 : ky_adim_max + 1, ky_adim_min:0],
            np.arange(nK2),
        )

    # def get_k_adim_loc(self):
    #     return self.get_k_adim()

    def get_dimX_K(self):
        return 0, 1, 2

    # def get_seq_indices_first_K(self):
    #     return 0, 0

    # def compute_energy_from_spatial(self, ff):
    #     return np.mean(abs(ff)**2)/2

    def project_fft_on_realX(self, ff_fft):
        return self.fft2d(self.ifft2d(ff_fft))

    def build_invariant_arrayX_from_2d_indices12X(self, o2d, arr2d):

        nX0, nX1, nX2 = self.get_shapeX_seq()
        nX0loc, nX1loc, nX2loc = self.get_shapeX_loc()

        if (nX1, nX2) != o2d.get_shapeX_seq():
            raise ValueError("Not the same physical shape...")

        # check that the 2d fft is not with distributed memory...
        if o2d.get_shapeX_loc() != o2d.get_shapeX_seq():
            raise ValueError("2d fft is with distributed memory...")

        ind0seq_first, ind1seq_first, _ = self.get_seq_indices_first_K()

        if (nX1loc, nX2loc) == o2d.get_shapeX_loc():
            arr3d_loc_2dslice = arr2d
        else:
            raise NotImplementedError

        arr3d = np.empty([nX0loc, nX1loc, nX2loc])
        for i0 in range(nX0loc):
            arr3d[i0] = arr3d_loc_2dslice

        return arr3d

    def build_invariant_arrayK_from_2d_indices12X(self, o2d, arr2d):

        nK0, nK1, nK2 = self.get_shapeK_seq()
        ret = np.zeros((nK0,) + o2d.shapeK_seq, dtype=np.complex128)

        ret[0] = arr2d
        return ret

    def get_seq_indices_first_X(self):
        """Get the "sequential" indices of the first number in Real space."""
        return 0, 0, 0

    def get_seq_indices_first_K(self):
        """Get the "sequential" indices of the first number in Fourier space."""
        return 0, 0, 0

    def get_k_adim_loc(self):
        """Get the non-dimensional wavenumbers stored locally.

        Returns
        -------

        k0_adim_loc : np.ndarray

        k1_adim_loc : np.ndarray

        k2_adim_loc :  np.ndarray

        The indices correspond to the index of the dimension in spectral space.

        """

        nK0, nK1, nK2 = self.get_shapeK_seq()
        nK0_loc, nK1_loc, nK2_loc = self.get_shapeK_loc()

        d0, d1, d2 = self.get_dimX_K()
        i0_start, i1_start, i2_start = self.get_seq_indices_first_K()

        k0_adim = compute_k_adim_seq_3d(nK0, d0)
        k0_adim_loc = k0_adim[i0_start : i0_start + nK0_loc]

        k1_adim = compute_k_adim_seq_3d(nK1, d1)
        k1_adim_loc = k1_adim[i1_start : i1_start + nK1_loc]

        k2_adim_loc = compute_k_adim_seq_3d(nK2, d2)

        return k0_adim_loc, k1_adim_loc, k2_adim_loc


def compute_k_adim_seq_3d(nk, axis):
    """Compute the adimensional wavenumber for an axis.

    Parameters
    ----------

    nk : int

      Global size in Fourier space for the axis.

    axis : int

      Index of the axis in real space (0 for z, 1 for y and 2 for x).

    """
    if axis == 2:
        return np.arange(nk)

    else:
        k_adim_max = nk // 2
        k_adim_min = -((nk - 1) // 2)
        return np.r_[0 : k_adim_max + 1, k_adim_min:0]


class FFTW1D(BasePyFFT):
    """ A class to use fftw 1D """

    def __init__(self, n):
        try:
            import pyfftw
        except ImportError as err:
            raise ImportError("ImportError. Instead fftpack?", err)

        if n % 2 != 0:
            raise ValueError("n should be even")

        shapeX = (n,)
        shapeK = (n,)
        self.shapeX = shapeX
        self.shapeK = self.shapeK_seq = self.shapeK_loc = shapeK
        self.arrayX = pyfftw.empty_aligned(shapeX, "complex128")
        self.arrayK = pyfftw.empty_aligned(shapeK, "complex128")
        self.fftplan = pyfftw.FFTW(
            input_array=self.arrayX,
            output_array=self.arrayK,
            axes=(-1,),
            direction="FFTW_FORWARD",
            threads=nthreads,
        )
        self.ifftplan = pyfftw.FFTW(
            input_array=self.arrayK,
            output_array=self.arrayX,
            axes=(-1,),
            direction="FFTW_BACKWARD",
            threads=nthreads,
        )

        self.coef_norm = n
        self.inv_coef_norm = 1.0 / n

    def fft(self, ff):
        self.arrayX[:] = ff
        self.fftplan()
        return self.arrayK / self.coef_norm

    def ifft(self, ff_fft):
        self.arrayK[:] = ff_fft
        self.ifftplan()
        return self.arrayX.copy()


class FFTW1DReal2Complex(BasePyFFT):
    """ A class to use fftw 1D """

    def __init__(self, arg, axis=-1):
        try:
            import pyfftw
        except ImportError as err:
            raise ImportError("ImportError. Instead fftpack?", err)

        if isinstance(arg, int):
            n = arg
            shapeX = (n,)
            shapeK = (n // 2 + 1,)
        else:
            n = arg[axis]
            shapeX = arg
            shapeK = list(copy(arg))
            shapeK[axis] = n // 2 + 1
            shapeK = tuple(shapeK)

        if n % 2 != 0:
            raise ValueError("n should be even")

        self.shapeX = shapeX
        self.shapeK = shapeK
        self.arrayX = pyfftw.empty_aligned(shapeX, "float64")
        self.arrayK = pyfftw.empty_aligned(shapeK, "complex128")
        self.fftplan = pyfftw.FFTW(
            input_array=self.arrayX,
            output_array=self.arrayK,
            axes=(axis,),
            direction="FFTW_FORWARD",
            threads=nthreads,
        )
        self.ifftplan = pyfftw.FFTW(
            input_array=self.arrayK,
            output_array=self.arrayX,
            axes=(axis,),
            direction="FFTW_BACKWARD",
            threads=nthreads,
        )

        self.coef_norm = n
        self.inv_coef_norm = 1.0 / n

    def fft(self, ff):
        self.arrayX[:] = ff
        self.fftplan(normalise_idft=False)
        return self.arrayK / self.coef_norm

    def ifft(self, ff_fft):
        self.arrayK[:] = ff_fft
        self.ifftplan(normalise_idft=False)
        return self.arrayX.copy()

    def sum_wavenumbers(self, ff_fft):
        if self.shapeX[0] % 2 == 0:
            return ff_fft[0] + ff_fft[-1] + 2 * np.sum(ff_fft[1:-1])

        else:
            return ff_fft[0] + 2 * np.sum(ff_fft[1:])

    def compute_energy_from_Fourier(self, ff_fft):
        return (
            abs(ff_fft[0]) ** 2
            + 2 * np.sum(abs(ff_fft[1:-1]) ** 2)
            + abs(ff_fft[-1]) ** 2
        ) / 2

    def compute_energy_from_spatial(self, ff):
        return np.mean(abs(ff) ** 2) / 2



"""
Utilities
=========

.. autosummary::
   :toctree:

   util
   constants
   mpi
   paramcontainer
   paramcontainer_gui
   serieofarrays
   timer
   daemons
   logger
   userconfig
   terminal_colors
   matlab2py
   info
   mail
   xmltotext

"""

from .util import (
    Params,
    time_as_str,
    config_logging,
    create_object_from_file,
    is_run_from_ipython,
    get_memory_usage,
    print_memory_usage,
    import_class,
    is_run_from_jupyter,
    modification_date,
)

__all__ = [
    "Params",
    "time_as_str",
    "config_logging",
    "create_object_from_file",
    "is_run_from_ipython",
    "get_memory_usage",
    "print_memory_usage",
    "import_class",
    "is_run_from_jupyter",
    "modification_date",
]



"""Handle ipython notebooks (:mod:`fluiddoc.ipynb_maker`)
=========================================================

.. autofunction:: ipynb_to_rst

"""

import os
from glob import glob
from datetime import datetime
import subprocess


def modification_date(filename):
    t = os.path.getmtime(filename)
    return datetime.fromtimestamp(t)


def call_bash(commands):
    subprocess.call(["/bin/bash", "-c", commands])


def ipynb_to_rst(path="ipynb", executed=None):
    """Convert notebooks to rst files

    If the user does not specify that the notebooks have already been executed,
    they are executed with jupyter-nbconvert before the conversion to rst
    files.

    """

    paths_ipynb = glob(path + "/*.ipynb")
    paths_ipynb = [
        path for path in paths_ipynb if not path.endswith(".nbconvert.ipynb")
    ]

    paths_ipynb_executed = []

    for filepath in paths_ipynb:
        if executed:
            try:
                executed[0]
            except TypeError:
                paths_ipynb_executed.append(filepath)
                continue

            else:
                nfile = os.path.split(filepath)[-1]
                if nfile in executed:
                    paths_ipynb_executed.append(filepath)
                    continue

        basename = os.path.splitext(filepath)[0]
        ipynb_executed = basename + ".nbconvert.ipynb"
        paths_ipynb_executed.append(ipynb_executed)

        if not os.path.exists(ipynb_executed) or modification_date(
            filepath
        ) > modification_date(ipynb_executed):
            call_bash(
                "jupyter-nbconvert --ExecutePreprocessor.timeout=200 "
                "--to notebook --execute " + filepath
            )

    for filepath in paths_ipynb_executed:
        basename = os.path.splitext(os.path.splitext(filepath)[0])[0]
        rstpath = basename + ".rst"
        rstname = os.path.split(rstpath)[-1]

        if not os.path.exists(rstpath):
            has_to_be_compiled = True
        else:
            d_ipynb = modification_date(filepath)
            d_rst = modification_date(rstpath)
            if d_ipynb > d_rst:
                has_to_be_compiled = True
            else:
                has_to_be_compiled = False

        if has_to_be_compiled:
            call_bash(
                "jupyter-nbconvert --to rst " + filepath + " --output " + rstname
            )



"""
Test txt functions
==================

"""

import unittest
import os
from shutil import rmtree

import numpy as np

from ..binary import BinFile


class TestBinary(unittest.TestCase):
    """Test fluiddyn.io.binary module."""

    def setUp(self):
        self._work_dir = "test_fluiddyn_io_binary"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        os.chdir(self._work_dir)

    def tearDown(self):
        os.chdir("..")
        rmtree(self._work_dir)

    def test_dict(self):
        name = "myfile"
        s = "poum"
        l = [1, 3]
        a = np.array([1.0, 1.5])

        with BinFile(name, "w") as f:
            f.write_as(s, buffersize=1)
            f.write_as(l, "I", buffersize=1)
            f.write_as(a, "float64", buffersize=1)

        with BinFile(name) as f:
            s1 = f.readt(4, "s")
            l1 = f.readt(2, "I")
            a1 = f.readt(2, "float64")

        self.assertEqual(s.encode(), s1)
        self.assertEqual([int(n) for n in l], list(l1))
        self.assertTrue(np.allclose(a, np.array(a1)))


if __name__ == "__main__":
    unittest.main()



"""
Slurm clusters under SNIC (:mod:`fluiddyn.clusters.snic`)
=========================================================

.. currentmodule:: fluiddyn.clusters.snic

Provides:

.. autoclass:: Beskow
   :members:

.. autoclass:: Beskow32
   :members:

.. autoclass:: Beskow36
   :members:

.. autoclass:: Tetralith
   :members:

.. autoclass:: Abisko
   :members:

.. autoclass:: Kebnekaise
   :members:

"""
from __future__ import print_function

from os import getenv
from .slurm import ClusterSlurm


_venv = getenv("VIRTUAL_ENV", getenv("CONDA_PREFIX", getenv("LOCAL_PYTHON")))


if _venv is None:
    from warnings import warn

    warn(
        "Cannot detect a virtualenv / conda env. You should set an environment "
        "variable LOCAL_PYTHON instead for fluiddyn.clusters.snic to work."
    )


class Beskow(ClusterSlurm):
    name_cluster = "beskow"
    nb_cores_per_node = 32
    cmd_run_interactive = "aprun"
    max_walltime = "23:59:59"

    def __init__(self):
        super().__init__()
        self.check_name_cluster("SNIC_RESOURCE")
        self.commands_setting_env = [
            "source /etc/profile",
            "module load gcc/4.9.1",
            "module swap PrgEnv-cray PrgEnv-intel",
            "module swap intel intel/18.0.0.128",
            "module load cray-fftw/3.3.6.2",
            "export CRAY_ROOTFS=DSL",
            f"conda activate {_venv}",
        ]

        self.commands_unsetting_env = ["conda deactivate"]


class Beskow32(Beskow):
    constraint = "Haswell"


class Beskow36(Beskow):
    nb_cores_per_node = 36
    constraint = "Broadwell"


class Tetralith(ClusterSlurm):
    name_cluster = "tetralith"
    nb_cores_per_node = 32
    cmd_run = "mpprun"
    cmd_run_interactive = "mpirun"
    max_walltime = "7-00:00:00"

    def __init__(self):
        super().__init__()
        self.check_name_cluster("SNIC_RESOURCE")
        self.commands_setting_env = []

        # NOTE: Typically load the following modules and save them
        # Python/3.6.3-anaconda-5.0.1-nsc1 intel/2018a
        # buildtool-easybuild/3.5.3-nsc17d8ce4 buildenv-intel/2018a-eb
        # FFTW/3.3.6-nsc1
        self.commands_setting_env.extend(
            ["ml restore", f"source activate {_venv}"]
        )

        self.commands_unsetting_env = ["source deactivate"]


class Abisko(ClusterSlurm):
    name_cluster = "abisko"
    nb_cores_per_node = 24
    cmd_run_interactive = "mpirun"
    max_walltime = "7-00:00:00"

    def __init__(self):
        super().__init__()
        self.check_name_cluster("SNIC_RESOURCE")

        self.commands_setting_env = [
            "source /etc/profile",
            "module load GCC/6.3.0-2.27 OpenMPI/2.0.2",
            "module load HDF5/1.10.0-patch1",
            "module load FFTW/3.3.6",
            f"source {_venv}/bin/activate",
        ]

        self.commands_unsetting_env = []


class Kebnekaise(ClusterSlurm):
    name_cluster = "kebnekaise"
    nb_cores_per_node = 28
    cmd_run_interactive = "mpirun"
    max_walltime = "7-00:00:00"

    def __init__(self):
        super().__init__()
        self.check_name_cluster("SNIC_RESOURCE")
        self.commands_setting_env = ["source /etc/profile"]

        self.commands_setting_env.extend(
            [
                "module load foss/2017a",
                # also loads GCC/6.3.0-2.27  OpenMPI/2.0.2
                # OpenBLAS/0.2.19-LAPACK-3.7.0 FFTW/3.3.6
                "module rm FFTW/3.3.6",
                "module load HDF5/1.10.0-patch1",
                "module load Python/3.6.1",
                f"source {_venv}/bin/activate",
            ]
        )

        self.commands_unsetting_env = []


_host = getenv("SNIC_RESOURCE")
if _host == "beskow":
    ClusterSNIC = Beskow
elif _host == "tetralith":
    ClusterSNIC = Tetralith
elif _host == "abisko":
    ClusterSNIC = Abisko
elif _host == "kebnekaise":
    ClusterSNIC = Kebnekaise


if __name__ == "__main__":
    cluster = ClusterSNIC()
    print("\n".join(cluster.commands_setting_env))



"""
Test query module
=================

"""

import builtins

# import sys
from contextlib import contextmanager
import unittest

import subprocess

from .. import query
from ...io.redirect_stdout import stdout_redirected


def do_nothing(*args, **kwargs):
    pass


@contextmanager
def mock_input(mock):
    # mocking write works only in Python 3
    # original_write = sys.stdout.write
    original_input = builtins.input
    # sys.stdout.write = do_nothing
    builtins.input = lambda: mock
    yield

    # sys.stdout.write = original_write
    builtins.input = original_input


@contextmanager
def mock_call():
    original_call = subprocess.call
    subprocess.call = do_nothing
    yield

    subprocess.call = original_call


class TestQuery(unittest.TestCase):
    """Test fluiddyn.io.query module."""

    def test_query(self):
        with mock_input("1.2"):
            self.assertEqual(1.2, query.query_number(""))

        with mock_input("test"):
            with stdout_redirected():
                query.query("?", "t")

        with mock_input("yes"):
            with mock_call():
                with stdout_redirected():
                    query.run_asking_agreement("ls")


if __name__ == "__main__":
    unittest.main()



"""Variable container (:mod:`fluiddyn.calcul.setofvariables`)
=============================================================

Provides:

.. autoclass:: SetOfVariables
   :members:
   :private-members:

"""

from __future__ import print_function

import numpy as np


class SetOfVariables(np.ndarray):
    """np.ndarray containing the variables.

    Parameters
    ----------

    input_array : :class:`numpy.ndarray`, optional
        Input array, default to None.

    keys : Iterable, optional
        Keys corresponding to the variables, default to None.

    shape_variable : Iterable, optional
        Shape of an array containing one variable, default to None.

    like : :class:`SetOfVariables`, optional
        Model for creating the new :class:`SetOfVariables`, default to None.

    value : Number, optional
        For initialization of the new :class:`SetOfVariables`, default to None.

    info : :class:`str`, optional
        Description, default to None.

    dtype : :class:`numpy.dtype`, optional
        see :class:`numpy.ndarray` help.

    """

    def __new__(
        cls,
        input_array=None,
        keys=None,
        shape_variable=None,
        like=None,
        value=None,
        info=None,
        dtype=None,
        **kargs
    ):
        if input_array is not None:
            arr = input_array

            if keys is None:
                raise ValueError(
                    "If input_array is provided, keys has to be provided."
                )

            if len(keys) != arr.shape[0]:
                raise ValueError(
                    "len(keys) has to be equal to input_array.shape[0]"
                )

        else:
            if like is not None:
                info = like.info
                keys = like.keys
                shape = like.shape
                dtype = like.dtype
            elif keys is None or shape_variable is None:
                raise ValueError(
                    "If like is not provided, keys and "
                    "shape_variable should be provided"
                )

            else:
                shape = (len(keys),) + tuple(shape_variable)

            arr = np.empty(shape, dtype=dtype)
            if value is not None:
                arr.fill(value)

        obj = np.asarray(arr, dtype=dtype, **kargs).view(cls)

        obj.info = info
        obj.keys = tuple(keys)
        obj.nvar = len(keys)

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

        if obj.shape == self.shape:
            self.info = getattr(obj, "info", None)
            self.keys = getattr(obj, "keys", None)
            self.nvar = getattr(obj, "nvar", None)

    def set_var(self, arg, value):
        """Set a variable."""
        if isinstance(arg, int):
            index = arg
        else:
            try:
                index = self.keys.index(arg)
            except ValueError:
                raise ValueError(
                    "'" + arg + "' is not in tuple " + repr(self.keys)
                )

        # warning: copy... costly!
        self[index] = value

    def get_var(self, arg):
        """Get a variable as a np.array."""
        if isinstance(arg, int):
            index = arg
        else:
            try:
                index = self.keys.index(arg)
            except ValueError:
                raise ValueError(
                    "'" + arg + "' is not in tuple " + repr(self.keys)
                )

        return np.asarray(self[index])

    def initialize(self, value=0):
        """Initialize as a constant array."""
        self[:] = value


if __name__ == "__main__":

    # shape_var = (2, 2)
    # keys = ['a', 'b']
    # shape = (len(keys),) + shape_var
    # data = np.arange(reduce(lambda x, y: x*y, shape)).reshape(shape)

    # a = SetOfVariables(data, dtype=None,
    #                   info='poum', keys=keys)

    # print('a:', a)
    # b = 2*a
    # print('b:', b)

    # c = SetOfVariables(like=b, value=2.)
    # print('c:', c)

    # print(type(a*c))
    # print(type(2*c))
    # print(type(c.get_var('a')))

    sov = SetOfVariables(
        keys=["rot_fft"],
        shape_variable=(33, 18),
        dtype=np.complex128,
        info="state_spect",
    )



"""
OAR clusters (:mod:`fluiddyn.clusters.oar`)
===========================================

Provides:

.. autoclass:: ClusterOAR
   :members:

"""

from __future__ import print_function

from builtins import str

import os
import datetime
import stat
from sys import version_info as version

from . import subprocess, Cluster
from ..io.query import run_asking_agreement, call_bash


class ClusterOAR(Cluster):
    name_cluster = ""
    nb_cores_per_node = 12
    has_to_add_name_cluster = False

    _doc_commands = """
Useful commands
---------------
oarsub -S script.sh
oarstat -u
oardel $JOB_ID
oarsub -C $JOB_ID"""

    def __init__(self):
        self.commands_setting_env = [
            "source /etc/profile",
            "module load python/{}.{}.{}".format(
                version.major, version.minor, version.micro
            ),
            "source /home/users/$USER/opt/mypy{}.{}/bin/activate".format(
                version.major, version.minor
            ),
        ]

    def check_oar(self):
        """check if this script is run on a frontal with oar installed"""
        try:
            subprocess.check_call(["oarsub", "--version"], stdout=subprocess.PIPE)
        except OSError:
            raise OSError("oar does not seem to be installed.")

    def submit_script(
        self,
        path,
        name_run="fluiddyn",
        nb_nodes=1,
        nb_cores_per_node=1,
        walltime="24:00:00",
        project=None,
        nb_mpi_processes=None,
        omp_num_threads=None,
        idempotent=False,
        delay_signal_walltime=300,
        network_address=None,
        ask=True,
        submit=True,
    ):

        path = os.path.expanduser(path)
        if not os.path.exists(path.split(" ")[0]):
            raise ValueError("The script does not exists! path:\n" + path)

        if not path.startswith("python "):
            command = "python " + path

        self.submit_command(
            command,
            name_run=name_run,
            nb_nodes=nb_nodes,
            nb_cores_per_node=nb_cores_per_node,
            walltime=walltime,
            project=project,
            nb_mpi_processes=nb_mpi_processes,
            omp_num_threads=omp_num_threads,
            idempotent=idempotent,
            delay_signal_walltime=delay_signal_walltime,
            network_address=network_address,
            ask=ask,
            submit=submit,
        )

    def submit_command(
        self,
        command,
        name_run="fluiddyn",
        nb_nodes=1,
        nb_cores_per_node=1,
        walltime="24:00:00",
        project=None,
        nb_mpi_processes=None,
        omp_num_threads=None,
        idempotent=False,
        delay_signal_walltime=300,
        network_address=None,
        ask=True,
        submit=True,
    ):

        self.check_oar()

        if nb_cores_per_node is None and nb_mpi_processes is not None:
            nb_cores_per_node = nb_mpi_processes

        if nb_cores_per_node > self.nb_cores_per_node:
            raise ValueError("Too many cores...")

        str_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        path_launching_script = "oar_launcher_" + str_time
        if os.path.exists(path_launching_script):
            i = 1
            while os.path.exists(path_launching_script + "_" + str(i)):
                i += 1
            path_launching_script += "_" + str(i)

        txt = self._create_txt_launching_script(
            command,
            name_run,
            nb_nodes,
            nb_cores_per_node,
            walltime,
            nb_mpi_processes=nb_mpi_processes,
            omp_num_threads=omp_num_threads,
            network_address=network_address,
        )

        with open(path_launching_script, "w") as f:
            f.write(txt)

        os.chmod(
            path_launching_script, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
        )

        launching_command = "oarsub"

        if project is not None:
            launching_command += " --project " + project

        if delay_signal_walltime is not None:
            launching_command += " --checkpoint " + str(delay_signal_walltime)

        if idempotent:
            launching_command += " -t idempotent"

        launching_command += " -S ./" + path_launching_script

        print(f'A launcher for the command "{command}" has been created.')
        if submit:
            if ask:
                run_asking_agreement(launching_command)
            else:
                print(
                    "The script is submitted with the command:\n",
                    launching_command,
                )
                call_bash(launching_command)

    def _create_txt_launching_script(
        self,
        command,
        name_run,
        nb_nodes,
        nb_cores_per_node,
        walltime,
        nb_mpi_processes=None,
        omp_num_threads=None,
        network_address=None,
    ):

        txt = "#!/bin/bash\n\n" "#OAR -n {}\n".format(name_run)

        txt += "#OAR -l "

        if self.has_to_add_name_cluster and network_address is None:
            txt += "{cluster='" + self.name_cluster + "'}"
        elif network_address is not None:
            txt += "{network_address='" + network_address + "'}"

        txt += "/nodes={}/core={},walltime={}\n\n".format(
            nb_nodes, nb_cores_per_node, walltime
        )

        txt += 'echo "hostname: "$HOSTNAME\n\n'

        txt += "\n".join(self.commands_setting_env) + "\n\n"

        if omp_num_threads is not None:
            txt += f"export OMP_NUM_THREADS={omp_num_threads}\n\n"

        if nb_mpi_processes is not None and nb_nodes > 1:
            txt += (
                "# Shell with environment variables forwarded\n"
                "export OMPI_MCA_plm_rsh_agent=oar-envsh\n\n"
            )

        txt += "exec "

        if nb_mpi_processes is not None:
            txt += f"mpirun -np {nb_mpi_processes} "

            if nb_nodes > 1:
                txt += "-machinefile ${OAR_NODEFILE} "

        txt += command + "\n"

        return txt



"""Mail (:mod:`fluiddyn.util.mail`)
===================================

Provides:

.. autofunction:: send_email

"""

from __future__ import print_function

import os
import smtplib
import socket

# For guessing MIME type based on file name extension
import mimetypes

from email import encoders
from email.utils import formatdate

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage


def send_email(
    subject,
    txt,
    address_recipients,
    address_sender=None,
    server="localhost",
    files=None,
):
    """Send an email.

    Inspired by an example in
    https://docs.python.org/2/library/email-examples.html

    Parameters
    ----------

    subject : str or unicode

      The subject of the email.

    txt : {str, unicode}

      The text of the email.

    address_recipients : {str, list}

      The recipient(s) of the mail.

    address_sender : {None, str}

      The sender.

    server : {None, str}

      A string defining a SMTP server (for example 'localhost:25')

    files : list

      A list of files that have to be attached.

    Notes
    -----

    You may have to start a SMTP local server.

    """

    if isinstance(address_recipients, str):
        address_recipients = [address_recipients]

    assert isinstance(address_recipients, list)

    if address_sender is None:
        address_sender = address_recipients[0]

    msg = MIMEMultipart()
    msg["Subject"] = subject
    msg["From"] = address_sender
    msg["To"] = ", ".join(address_recipients)
    msg["Date"] = formatdate(localtime=True)

    # main part: the text
    part = MIMEText(txt)
    msg.attach(part)

    if files is None:
        files = []

    # sub parts: the enclosed files
    for path in files:
        # Guess the content type based on the file's extension.  Encoding
        # will be ignored, although we should check for simple things like
        # gzip'd or compressed files.
        ctype, encoding = mimetypes.guess_type(path)
        if ctype is None or encoding is not None:
            # No guess could be made, or the file is encoded
            # (compressed), so
            # use a generic bag-of-bits type.
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)
        with open(path, "rb") as fp:
            if maintype == "text":
                # Note: we should handle calculating the charset
                part = MIMEText(fp.read(), _subtype=subtype)
            elif maintype == "image":
                part = MIMEImage(fp.read(), _subtype=subtype)
            elif maintype == "audio":
                part = MIMEAudio(fp.read(), _subtype=subtype)
            else:
                part = MIMEBase(maintype, subtype)
                part.set_payload(fp.read())
                # Encode the payload using Base64
                encoders.encode_base64(part)
        # set the filename parameter
        part.add_header(
            "Content-Disposition", "attachment", filename=os.path.split(path)[1]
        )

        msg.attach(part)
    try:
        server = smtplib.SMTP(host=server)
    except socket.gaierror as e:
        print("socket.gaierror: you may have to setup a local SMTP server.")
        raise e

    server.sendmail(address_sender, address_recipients, msg.as_string())
    server.quit()


if __name__ == "__main__":

    send_email("email test", "blablabla\n" * 3, "pierre.augier@legi.cnrs.fr")




def gradient_colors(nb_colors, color_start=None, color_end=None):
    """Produce a color gradient."""
    if color_start is None:
        color_start = [1, 0, 0]
    if color_end is None:
        color_end = [0, 0, 1]
    # start at black, finish at white
    gradient = [color_start]
    # If only one color, return black
    if nb_colors == 1:
        return gradient

    # Calcuate a color at each evenly spaced value
    # of t = i / n from i in 0 to 1
    for t in range(1, nb_colors):
        gradient.append(
            [
                color_start[j]
                + (float(t) / (nb_colors - 1)) * (color_end[j] - color_start[j])
                for j in range(3)
            ]
        )
    return gradient



print("import: 'fluiddoc'")
import fluiddoc

print("import: 'fluiddyn'")
import fluiddyn

print("import: 'fluiddyn.calcul'")
import fluiddyn.calcul

print("import: 'fluiddyn.calcul.test'")
import fluiddyn.calcul.test

print("import: 'fluiddyn.clusters'")
import fluiddyn.clusters

print("import: 'fluiddyn.clusters.test'")
import fluiddyn.clusters.test

print("import: 'fluiddyn.io'")
import fluiddyn.io

print("import: 'fluiddyn.io.test'")
import fluiddyn.io.test

print("import: 'fluiddyn.output'")
import fluiddyn.output

print("import: 'fluiddyn.output.test'")
import fluiddyn.output.test

print("import: 'fluiddyn.util'")
import fluiddyn.util

print("import: 'fluiddyn.util.matlab2py'")
import fluiddyn.util.matlab2py

print("import: 'fluiddyn.util.matlab2py.test'")
import fluiddyn.util.matlab2py.test

print("import: 'fluiddyn.util.test'")
import fluiddyn.util.test




r"""Sphinx extension providing a new directive ``mathmacro``
============================================================

This extension has to be added after the other math extension since it
redefined the math directive and the math role. For example, like this
(in the conf.py file)::

  extensions = [
      'sphinx.ext.autodoc', 'sphinx.ext.doctest',
      'sphinx.ext.mathjax',
      'sphinx.ext.viewcode', 'sphinx.ext.autosummary',
      'numpydoc',
      'fluiddoc.mathmacro']


Here, I show how to use a new mathmacro substitution directive in
reStructuredText. I think even this small example demonstrates that it
is useful.

Example
-------

First some math without math macros.  Let's take the example of the incompressible
Navier-Stokes equation:

.. math::
  \mbox{D}_t \textbf{v} =
  -\boldsymbol{\nabla} p + \nu \boldsymbol{\nabla} ^2 \textbf{v}.

where :math:`\mbox{D}_t` is the convective derivative, :math:`\textbf{v}` the
velocity, :math:`\boldsymbol{\nabla}` the nabla operator, :math:`\nu` the
viscosity and :math:`\boldsymbol{\nabla}^2` the Laplacian operator.

The code for this is:

.. code-block:: rest

  .. math::
    \mbox{D}_t \textbf{v} =
    -\boldsymbol{\nabla} p + \nu \boldsymbol{\nabla} ^2 \textbf{v}.

  where :math:`\mbox{D}_t` is the convective derivative, :math:`\textbf{v}` the
  velocity, :math:`\boldsymbol{\nabla}` the nabla operator, :math:`\nu` the
  viscosity and :math:`\boldsymbol{\nabla}^2` the Laplacian operator.


.. |Dt| mathmacro:: \mbox{D}_t
.. |bnabla| mathmacro:: \boldsymbol{\nabla}
.. |vv| mathmacro:: \textbf{v}

Now, let's use some math macros and try to get the same result... We use the
following code:

.. code-block:: rest

  The Navier-Stokes equation can now be written like this:

  .. math:: \Dt \vv = - \bnabla p + \nu \bnabla^2 \vv

  where |Dt| is the convective derivative, |vv| the velocity, |bnabla| the nabla
  operator, :math:`\nu` the viscosity and :math:`\bnabla^2` the Laplacian operator.

It gives:

The Navier-Stokes equation can now be written like this:

.. math:: \Dt \vv = - \bnabla p + \nu \bnabla^2 \vv

where |Dt| is the convective derivative, |vv| the velocity, |bnabla| the nabla
operator, :math:`\nu` the viscosity and :math:`\bnabla^2` the Laplacian operator.

"""

import re

from docutils.parsers.rst.directives.misc import Replace
from docutils.parsers.rst.roles import math_role as old_math_role

from sphinx.directives.patches import MathDirective


def multiple_replacer(replace_dict):
    """Return a function replacing doing multiple replacements.

    The produced function replace `replace_dict.keys()` by
    `replace_dict.values`, respectively.

    """

    def replacement_function(match):
        s = match.group(0)
        end = s[-1]
        if re.match(r"[\W_]", end):
            return replace_dict[s[:-1]] + end
        else:
            return replace_dict[s]

    pattern = "|".join(
        [re.escape(k) + r"[\W_$]" for k in replace_dict.keys()]
        + [re.escape(k) + "$" for k in replace_dict.keys()]
    )
    pattern = re.compile(pattern, re.M)

    def _mreplace(string):
        return pattern.sub(replacement_function, string)

    return _mreplace


def multiple_replace(string, replace_dict):
    mreplace = multiple_replacer(replace_dict)
    return mreplace(string)


class MathMacro(Replace):
    """Directive defining a math macro."""

    def run(self):
        if not hasattr(self.state.document, "math_macros"):
            self.state.document.math_macros = {}

        latex_key = "\\" + self.state.parent.rawsource.split("|")[1]
        self.state.document.math_macros[latex_key] = "".join(self.content)

        self.content[0] = ":math:`" + self.content[0]
        self.content[-1] = self.content[-1] + "`"

        return super().run()


class NewMathDirective(MathDirective):
    """New math block directive parsing the latex code."""

    def run(self):
        try:
            math_macros = self.state.document.math_macros
        except AttributeError:
            pass
        else:
            if math_macros:
                mreplace = multiple_replacer(math_macros)
                for i, c in enumerate(self.content):
                    self.content[i] = mreplace(c)
                for i, a in enumerate(self.arguments):
                    self.arguments[i] = mreplace(a)
        return super().run()


def new_math_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """New math role parsing the latex code."""
    try:
        math_macros = inliner.document.math_macros
    except AttributeError:
        pass
    else:
        if math_macros:
            rawtext = multiple_replace(rawtext, math_macros)
            text = rawtext.split("`")[1]

    return old_math_role(
        role, rawtext, text, lineno, inliner, options=options, content=content
    )


def setup(app):

    app.add_role("math", new_math_role, override=True)
    app.add_directive("math", NewMathDirective, override=True)

    app.add_directive("mathmacro", MathMacro)


if __name__ == "__main__":
    math_macros = {"\\vv": "\\textbf{v}"}
    s = "\\vv"
    mreplace = multiple_replacer(math_macros)
    print(mreplace(s))



from __future__ import print_function
import unittest
import os
from shutil import rmtree
import json

from ..paramcontainer import ParamContainer, tidy_container
from ...io.redirect_stdout import stdout_redirected

xml_file = os.path.abspath(os.path.join(os.path.dirname(__file__), "file.xml"))


class TestContainer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.params = ParamContainer(tag="params")
        params = cls.params
        params._set_attrib("a0", 1)
        params._set_attribs({"a1": 1, "a2": "a", "a_str_list": ["a", "b"]})

        params._set_child("child0", {"a0": 2, "a1": None})
        params.child0.a0 = []
        params.child0.a0.append(1)

        params.child0._set_child("cc", {})

        params.child0.cc._set_doc("Hello")

        with stdout_redirected():
            params._print_docs()

        params._get_key_attribs()

        cls.txt = params._make_xml_text()
        cls._work_dir = "test_fluiddyn_util_paramcontainer"
        if not os.path.exists(cls._work_dir):
            os.mkdir(cls._work_dir)

        os.chdir(cls._work_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test_save_load_xml(self):
        """Test save to and load from xml file."""
        params = self.params

        name_file_xml = "params_test.xml"
        params._save_as_xml(name_file_xml)
        params._save_as_xml(comment="Hello")

        params2 = ParamContainer(path_file=name_file_xml)
        os.remove(name_file_xml)
        params2._make_xml_text()

        self.assertEqual(self.params, params2)

        params2["a1"]
        params2["a1"] = 2
        with stdout_redirected():
            params2._print_doc()
            params2._print_as_xml()
            repr(params2)

        params3 = ParamContainer(tag="params3")
        params3._set_attrib("a0", 1)
        params2._set_as_child(params3)

    def test_save_load_hdf5(self):
        """Test save to and load from hdf5 file."""
        params = self.params

        name_file_h5 = "params_test.h5"
        params._save_as_hdf5(path_file=name_file_h5)
        params._save_as_hdf5()

        params2 = ParamContainer(path_file=name_file_h5)
        os.remove(name_file_h5)

        try:
            self.assertEqual(self.params, params2)
        except AssertionError:
            print("params=\n", self.params)
            print("params2=\n", params2)
            raise

    def test_raise(self):
        """Test raise errors."""
        params = self.params

        with self.assertRaises(AttributeError):
            params.does_not_exist = 1

        with self.assertRaises(AttributeError):
            params._does_not_exist

        with self.assertRaises(AttributeError):
            print(params.child1.a0)

    def test_tidy(self):
        param = ParamContainer(path_file=xml_file)
        param._make_xml_text()
        tidy_container(param)

    def test_render_json(self):
        """Test JSON rendering (used in Jupyter)."""
        data, metadata = self.params._repr_json_()
        json.dumps(data)


if __name__ == "__main__":
    unittest.main()



"""
Test digiflow module
====================

"""

import unittest
import os
from shutil import rmtree

# import numpy as np

from ..digiflow import DigiflowImage, DigiflowMovie


class TestDigiflow(unittest.TestCase):
    """Test fluiddyn.io.digiflow module."""

    def setUp(self):
        self._work_dir = "test_fluiddyn_io_digiflow"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        os.chdir(self._work_dir)

        self.path_dfi = "myfile.dfi"
        self.path_dfm = "myfile.dfm"

    def tearDown(self):
        os.chdir("..")
        rmtree(self._work_dir)


if __name__ == "__main__":
    unittest.main()



"""
Local workstations as clusters (:mod:`fluiddyn.clusters.local`)
===============================================================

Provides:

.. autoclass:: ClusterLocal
   :members:

"""
from __future__ import print_function
from builtins import str

import os
import psutil
import datetime
import stat
from shlex import split
from socket import gethostname
from . import subprocess, Cluster
from ..io.query import run_asking_agreement, call_bash
from ..util.timer import TimeStr, timestamp_to_seconds


def is_python_script(path):
    """Check whether path points to a file ending with .py extension."""
    return os.path.splitext(path)[1] == ".py"


class ClusterLocal(Cluster):
    """Works as a compatibility layer to execute submit scripts on local
    workstations as clusters with no job schedulers.

    Instead uses POSIX commands:

     - ``nohup`` and ``mpirun`` to launch jobs.
     - ``timeout`` to set walltime..

    """

    _doc_commands = "\n".join(
        [
            "Useful commands",
            "---------------",
            "pgrep -af python",
            "killall python",
            "nohup python -u launcher.py >launcher.stdout 2>launcher.stderr &",
            "top",
        ]
    )
    name_cluster = gethostname()
    nb_cores_per_node = psutil.cpu_count()
    cmd_run = "mpirun"
    cmd_launch = "nohup"
    max_walltime = "30-00:00:00"

    def __init__(self):
        self.commands_setting_env = []
        self.commands_unsetting_env = []
        virtualenv = os.getenv("VIRTUAL_ENV")
        if virtualenv is not None:
            self.commands_setting_env.append(
                "source " + virtualenv + "/bin/activate"
            )
            self.commands_unsetting_env.append("deactivate")

    def submit_script(self, path, *args, **kwargs):
        """Submit a script. See `submit_command` for all possible arguments"""
        path = os.path.expandvars(path)
        script = path.split()[0]
        if not os.path.exists(script):
            raise ValueError("Script does not exist! path:\n" + script)

        if is_python_script(script) and not path.startswith("python "):
            path = "python " + path

        self.submit_command(path, *args, **kwargs)

    def submit_command(
        self,
        command,
        name_run="fluiddyn",
        nb_nodes=1,
        nb_cores_per_node=None,
        walltime="23:59:58",
        nb_mpi_processes=None,
        retain_script=True,
        omp_num_threads=1,
        ask=True,
        bash=True,
        interactive=False,
        **kwargs
    ):
        """Submit a command.

        Parameters
        ----------
        command : string
            Command which executes the run
        name_run : string
            Name of the run to be displayed in SLURM queue
        nb_nodes : integer
            Sets number of MPI processes = nb_nodes * nb_cores_per_node
        nb_cores_per_node : integer
            Defaults to a maximum is fixed for a cluster, as set by self.nb_cores_per_node.
        walltime : string
            Minimum walltime for the job
        nb_mpi_processes : integer
            Number of MPI processes. Defaults to a `nb_cores_per_node * nb_nodes`.
        omp_num_threads : integer
            Number of OpenMP threads
        ask : boolean
            Ask for user input to submit the jobscript or not
        bash : boolean
            Submit jobscript via `call_bash` function
        interactive : boolean
            When True do not use cmd_launch to execute.

        """
        nb_cores_per_node, nb_mpi_processes = self._parse_cores_procs(
            nb_nodes, nb_cores_per_node, nb_mpi_processes
        )

        path_launching_script = self._make_path_launching_script()
        self._check_walltime(walltime)

        create_txt_kwargs = locals()
        del create_txt_kwargs["self"]
        txt = self._create_txt_launching_script(**create_txt_kwargs)
        self._write_txt_launching_script(txt, path_launching_script)

        launching_command = " ./" + path_launching_script
        self._launch(launching_command, command, bash, ask)

        if not retain_script:
            os.remove(path_launching_script)

    def _parse_cores_procs(self, nb_nodes, nb_cores_per_node, nb_mpi_processes):
        """Parse number of cores per node and MPI processes when these are
        None.

        """
        if nb_cores_per_node is None:
            nb_cores_per_node = self.nb_cores_per_node
        elif nb_cores_per_node > self.nb_cores_per_node:
            raise ValueError("Too many cores...")

        if nb_mpi_processes is None:
            if nb_cores_per_node == 0:
                nb_mpi_processes = 1
            else:
                nb_mpi_processes = nb_cores_per_node * nb_nodes

        return nb_cores_per_node, nb_mpi_processes

    def _make_path_launching_script(
        self, prefix="launcher", path_launching_script=None
    ):
        """Make a uniques path and name of the launching script from the current
        time.

        """
        if path_launching_script is None:
            str_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            path_launching_script = prefix + "_" + str_time + ".sh"

        if os.path.exists(path_launching_script):
            i = 1
            while os.path.exists(path_launching_script + "_" + str(i)):
                i += 1
            path_launching_script += "_" + str(i)

        return path_launching_script

    def _check_walltime(self, walltime):
        """Check if walltime exceeds maximum walltime."""
        if walltime is None:
            return

        elif TimeStr(walltime) > self.max_walltime:
            raise ValueError(
                (
                    "Walltime requested {} exceeds permitted maximum walltime "
                    "{}"
                ).format(walltime, self.max_walltime)
            )

    def _launch(self, launching_command, command="", bash=True, ask=True):
        """Ask and launch the command using a subprocess call."""
        print(f"A launcher for the command {command} has been created.")

        if ask:
            run_asking_agreement(launching_command)
        else:
            print(
                "The script is submitted with the command:\n", launching_command
            )
            if bash:
                call_bash(launching_command)
            else:
                subprocess.call(split(launching_command))

    def _create_txt_launching_script(self, **kwargs):
        """Create the text for a script which launches the command."""
        path_launching_script = kwargs["path_launching_script"]
        command = kwargs["command"]
        name_run = kwargs["name_run"]
        nb_mpi_processes = kwargs["nb_mpi_processes"]
        walltime = kwargs["walltime"]
        omp_num_threads = kwargs["omp_num_threads"]
        interactive = kwargs["interactive"]

        logfile = f"LOCAL.{name_run}"
        logfile_stdout = logfile + ".$$.stdout"
        logfile_stderr = logfile + ".$$.stderr"

        txt = "#!/bin/bash -l\n\n"

        txt += 'echo "hostname: "$HOSTNAME\n\n'
        txt += self._log_job(
            nb_mpi_processes,
            path_launching_script,
            logfile_stdout,
            command,
            "LOCAL_JOB.md",
        )
        txt += "\n".join(self.commands_setting_env) + "\n\n"

        if omp_num_threads is not None:
            txt += f"export OMP_NUM_THREADS={omp_num_threads}\n\n"

        cmd = command
        if nb_mpi_processes > 1:
            cmd = f"{self.cmd_run} -n {nb_mpi_processes} {cmd}"

        walltime_seconds = timestamp_to_seconds(walltime)
        cmd = f"timeout -s TERM {walltime_seconds}s {cmd}"

        if interactive:
            txt += cmd
        else:
            txt += "{} {} >{} 2>{} &".format(
                self.cmd_launch, cmd, logfile_stdout, logfile_stderr
            )

        txt += "\n" + "\n".join(self.commands_unsetting_env)
        return txt

    def _log_job(
        self,
        nb_mpi_processes,
        path_launching_script,
        logfile_stdout,
        command,
        logfile_job,
    ):
        """Generate a shell command to log the job into a markdown file."""
        return (
            "\n" + r'printf "\n# np={} `date` PID $$ {} {}\n{}" >> {}' + "\n"
        ).format(
            nb_mpi_processes,
            path_launching_script,
            logfile_stdout,
            command,
            logfile_job,
        )

    def _write_txt_launching_script(self, txt, path_launching_script):
        """Write the text into an executable script which launches the command."""
        with open(path_launching_script, "w") as f:
            f.write(txt)

        os.chmod(
            path_launching_script, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
        )



"""Test fluidinfo
=================

"""
import unittest
import os
from ..info import _get_parser, main, print_sys_info
from ...io.redirect_stdout import stdout_redirected
import warnings


class TestInfo(unittest.TestCase):
    """Test fluiddyn.util.info module."""

    @classmethod
    def tearDownClass(cls):
        filename = "sys_info.xml"
        if os.path.exists(filename):
            os.remove(filename)

    def test_save_info(self):
        """Equivalent to testing command `fluidinfo -s`."""
        parser = _get_parser()
        args = parser.parse_args(["-s"])
        main(args)

    def test_print_info(self):
        """Equivalent to testing command `fluidinfo -v`."""
        # Warnings arise from third-party packages
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        warnings.filterwarnings("ignore", category=PendingDeprecationWarning)
        with stdout_redirected():
            print_sys_info(verbosity=1)

        warnings.resetwarnings()



"""User configuration (:mod:`fluiddyn.util.userconfig`)
=======================================================



"""

import os as _os
from runpy import run_path as _run_path


def load_user_conf_files(namepackage="fluiddyn", possible_conf_files=None):
    """Execute some user configuration files if they exist and gather the
    configuration values as module attributes.

    """

    conf_vars = {}

    home = _os.path.expanduser("~")

    possible_conf_files = [_os.path.join(home, "." + namepackage, "config.py")]

    conf_files = []
    for _path in possible_conf_files:
        if _os.path.isfile(_path):
            conf_files.append(_path)
            conf_vars = _run_path(_path, init_globals=conf_vars)

    conf_vars = {
        k: v for k, v in list(conf_vars.items()) if not k.startswith("__")
    }

    config = {k: v for k, v in list(conf_vars.items())}
    config["home"] = home
    config["possible_conf_files"] = possible_conf_files
    config["conf_files"] = conf_files
    config["conf_vars"] = conf_vars

    return config


if __name__ == "__main__":
    config = load_user_conf_files()

    glob = globals()
    for _k, _v in list(config.items()):
        glob[_k] = _v



import unittest

from ..setofvariables import SetOfVariables


class TestSetofvariables(unittest.TestCase):
    def test_sov(self):
        sov = SetOfVariables(keys=("a", "b", "c"), shape_variable=(2, 2))
        b = sov.get_var("b")
        sov.set_var("a", b)
        sov.initialize(3)

        sov1 = SetOfVariables(like=sov)



"""Redirect stdout (:mod:`fluiddyn.io.redirect_stdout`)
==========================================================


"""
from typing import Union, TextIO
import os
import io
import pathlib
import contextlib


__all__ = ["stdout_redirected"]


@contextlib.contextmanager
def stdout_redirected(
    doit: bool = True,
    to: Union[str, pathlib.Path, TextIO, io.StringIO] = os.devnull,
):
    """Redirect stdout to os.devnull or a buffer

    Parameters
    ----------
    doit : bool

        Toggle redirection.

    to : StringIO instance or file object or str

        Buffer or file or file path to redirect to.

    """
    if not doit:
        yield  # Control to the with block
        return  # Do nothing and exit contextmanager

    with contextlib.ExitStack() as stack:
        if isinstance(to, (str, pathlib.Path)):
            to_file = stack.enter_context(open(to, "w+"))
        elif isinstance(to, (io.TextIOBase, io.StringIO)):
            to_file = to

        stack.enter_context(contextlib.redirect_stdout(to_file))
        yield  # Control to the with block
        # Exit the contextmanagers and finally, exit the stack



"""
Slurm clusters (:mod:`fluiddyn.clusters.slurm`)
===============================================

Provides:

.. autoclass:: ClusterSlurm
   :members:

"""

import os

from . import subprocess
from .local import ClusterLocal


class ClusterSlurm(ClusterLocal):
    """Base class for clusters with SLURM job scheduler."""

    _doc_commands = """
Useful commands
---------------

sbatch
squeue -u $USER
scancel
scontrol hold
scontrol release"""
    name_cluster = ""
    nb_cores_per_node = 32
    default_project = None
    cmd_run = "srun"
    cmd_run_interactive = None
    cmd_launch = "sbatch"
    max_walltime = "23:59:59"

    def __init__(self):
        self.check_slurm()
        self.commands_setting_env = []
        self.commands_unsetting_env = []

    def check_slurm(self):
        """Check if this script is run on a frontal with slurm installed."""
        try:
            subprocess.check_call(["sbatch", "--version"], stdout=subprocess.PIPE)
            slurm_installed = True
        except OSError:
            slurm_installed = False

        if not slurm_installed:
            raise ValueError(
                "This script should be run on a cluster with slurm installed."
            )

    def check_name_cluster(self, env="HOSTNAME"):
        """Check if self.name_cluster matches the environment variable."""

        if self.name_cluster not in os.getenv(env):
            raise ValueError(
                "Cluster name mismatch detected; expected " + self.name_cluster
            )

    def submit_command(
        self,
        command,
        name_run="fluiddyn",
        nb_nodes=1,
        nb_cores_per_node=None,
        walltime="23:59:58",
        project=None,
        nb_mpi_processes=None,
        omp_num_threads=1,
        nb_runs=1,
        path_launching_script=None,
        path_resume=None,
        retain_script=True,
        jobid=None,
        requeue=False,
        nb_switches=None,
        max_waittime=None,
        ask=True,
        bash=True,
        email=None,
        interactive=False,
        signal_num=12,
        signal_time=300,
        **kwargs
    ):
        """Submit a command.

        Parameters
        ----------
        command : string
            Command which executes the run
        name_run : string
            Name of the run to be displayed in SLURM queue
        nb_nodes : integer
            Sets number of MPI processes = nb_nodes * nb_cores_per_node
        nb_cores_per_node : integer
            Defaults to a maximum is fixed for a cluster, as set by self.nb_cores_per_node.
            Set as 1 for a serial job. Set as 0 to spread jobs across nodes
            (starts job faster, maybe slower).
        walltime : string
            Minimum walltime for the job
        project : string
            Sets the allocation to run the job under
        nb_mpi_processes : integer
            Number of MPI processes. Defaults to a `nb_cores_per_node * nb_nodes`.
        omp_num_threads : integer
            Number of OpenMP threads
        nb_runs : integer
            Number of times to submit jobs (launch once using `command` and
            resume thereafter with `path_resume` script / command).
        path_launching_script: string
            Path of the SLURM jobscript
        path_resume : string
            Path of the script to resume a job, which takes one
            argument - the `path_run` parsed from the output.
        retain_script : boolean
            Retail or delete script after launching job
        jobid : integer
            Run under already allocated job
        requeue : boolean
            If set True, permit the job to be requeued.
        nb_switches : integer
            Max / Optimum switches
        max_waittime : string
            Max time to wait for optimum
        ask : boolean
            Ask for user input to submit the jobscript or not
        bash : boolean
            Submit jobscript via `call_bash` function
        email : string
            In case of failure notify to the specified email address
        interactive : boolean
            Use `cmd_run_interactive` instead of `cmd_run` inside the jobscript
        signal_num : int or False
        signal_time : int
            Send the signal `signal_num` `signal_time` seconds before the end of the job.
        """
        nb_cores_per_node, nb_mpi_processes = self._parse_cores_procs(
            nb_nodes, nb_cores_per_node, nb_mpi_processes
        )

        path_launching_script = self._make_path_launching_script()
        self._check_walltime(walltime)

        is_resume_script = bool("resumer" in path_launching_script)
        if project is None:
            project = self.default_project

        launching_command = self.cmd_launch

        if jobid is not None:
            launching_command += " --jobid=" + str(jobid)

        if requeue:
            launching_command += " --requeue"

        if is_resume_script:
            dependencies = input("Enter jobid dependencies :").split()
            launching_command += " --dependency=afternotok:" + ":".join(
                dependencies
            )
        else:
            dependencies = None

        if nb_switches is not None and max_waittime is not None:
            launching_command += (
                " --switches=" + str(nb_switches) + "{@" + max_waittime + "}"
            )

        create_txt_kwargs = locals()
        del create_txt_kwargs["self"]
        txt = self._create_txt_launching_script(**create_txt_kwargs)
        self._write_txt_launching_script(txt, path_launching_script)
        launching_command += " ./" + path_launching_script
        self._launch(launching_command, command, bash, ask)

        if not retain_script:
            os.remove(path_launching_script)

        nb_times_resume = int(nb_runs) - 1
        for n in range(0, nb_times_resume):
            nb_runs = 1
            jobid = None
            requeue = False
            path = path_resume
            path_launching_script = self._make_path_launching_script("resumer")
            submit_script_kwargs = locals()
            del submit_script_kwargs["self"]
            del submit_script_kwargs["command"]
            self.submit_script(**submit_script_kwargs)

    def _create_txt_launching_script(self, **kwargs):
        """
        Examples
        --------

            #!/bin/bash -l
            # The -l above is required to get the full environment with modules

            # Set the allocation to be charged for this job
            # not required if you have set a default allocation
            #SBATCH -A 201X-X-XX

            # The name of the script is myjob
            #SBATCH -J myjob

            # Only 1 hour wall-clock time will be given to this job
            #SBATCH -t 1:00:00

            # Number of nodes
            #SBATCH -N 4
            # Number of MPI processes per node (the following is actually the default)
            #SBATCH --ntasks-per-node=32
            # Number of MPI processes.
            #SBATCH -n 128

            #SBATCH -e error_file.e
            #SBATCH -o output_file.o

            # Run the executable named myexe
            srun -n 128 ./myexe
        """
        path_launching_script = kwargs["path_launching_script"]
        command = kwargs["command"]
        name_run = kwargs["name_run"]
        project = kwargs["project"]
        nb_nodes = kwargs["nb_nodes"]
        nb_cores_per_node = kwargs["nb_cores_per_node"]
        walltime = kwargs["walltime"]
        nb_mpi_processes = kwargs["nb_mpi_processes"]
        omp_num_threads = kwargs["omp_num_threads"]
        dependencies = kwargs["dependencies"]
        email = kwargs["email"]
        interactive = kwargs["interactive"]
        is_resume_script = kwargs["is_resume_script"]
        signal_num = kwargs["signal_num"]
        signal_time = kwargs["signal_time"]

        logfile = f"SLURM.{name_run}"
        logfile_stdout = logfile + ".${SLURM_JOBID}.stdout"

        txt = "#!/bin/bash -l\n\n"

        txt += f"#SBATCH -J {name_run}\n\n"
        if project is not None:
            txt += f"#SBATCH -A {project}\n\n"

        if signal_num:
            time = walltime
        else:
            time = self.max_walltime

        txt += f"#SBATCH --time={time}\n"
        txt += f"#SBATCH --time-min={walltime}\n"
        if signal_num:
            txt += f"#SBATCH --signal={signal_num}@{signal_time}\n"

        txt += f"#SBATCH --nodes={nb_nodes}\n"
        if nb_cores_per_node > 0:
            txt += f"#SBATCH --ntasks-per-node={nb_cores_per_node}\n"

        txt += f"#SBATCH --ntasks={nb_mpi_processes}\n\n"

        if hasattr(self, "constraint"):
            txt += "#SBATCH --constraint=" + self.constraint + "\n"

        if email is not None:
            txt += "#SBATCH --mail-type=FAIL\n"
            txt += f"#SBATCH --mail-user={email}\n"

        txt += f"#SBATCH -e {logfile}.%J.stderr\n"
        txt += f"#SBATCH -o {logfile}.%J.stdout\n\n"

        txt += 'echo "hostname: "$HOSTNAME\n\n'
        txt += self._log_job(
            nb_mpi_processes,
            path_launching_script,
            logfile_stdout,
            command,
            "SLURM_JOB.md",
        )

        txt += "\n".join(self.commands_setting_env) + "\n\n"

        if omp_num_threads is not None:
            txt += f"export OMP_NUM_THREADS={omp_num_threads}\n\n"

        if is_resume_script:
            jobid = dependencies[0]
            main_logfile = f"SLURM.{name_run}.{jobid}.stdout"
            txt += "PATH_RUN=$(sed -n '/path_run/{n;p;q}' " + "{})\n".format(
                main_logfile
            )

        if interactive:
            cmd = self.cmd_run_interactive
        else:
            cmd = self.cmd_run

        if nb_mpi_processes > 1:
            txt += f"{cmd} -n {nb_mpi_processes} "

        if is_resume_script:
            txt += f"{command} $PATH_RUN"
        else:
            txt += command

        if interactive:
            txt += f" > {logfile_stdout} 2>&1"

        txt += "\n" + "\n".join(self.commands_unsetting_env)
        return txt



import unittest
import os
from shutil import rmtree

from .. import figs
from ..figs import Figures, show
from ..util import gradient_colors
from ...io.redirect_stdout import stdout_redirected


def do_nothing(*args):
    pass


figs.plt.show = do_nothing
figs.plt.ion = do_nothing
figs.plt.ioff = do_nothing


class TestFigs(unittest.TestCase):
    """Test fluiddyn.output.figs module."""

    def setUp(self):
        self._work_dir = "test_fluiddyn_output_figs"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        self.figures = Figures(self._work_dir, hastosave=True, for_article=False)

    def tearDown(self):
        rmtree(self._work_dir)

    def test_gradient_colors(self):
        gradient_colors(4)
        gradient_colors(5)

    def test_save(self):

        fig = self.figures.new_figure(
            "figure0.png", fig_width_mm=40, fig_height_mm=40
        )
        with stdout_redirected():
            fig.saveifhasto()
        show()


if __name__ == "__main__":
    unittest.main(exit=True)



"""
Test ns3d module
====================

"""

import unittest
import os
from glob import glob
from shutil import rmtree, copy

from ..ns3d import NS3DFieldFile, NS3DForcingInfoFile
from ..redirect_stdout import stdout_redirected

input_dir = os.path.join(os.path.dirname(__file__), "ns3d_files")
input_files = glob(os.path.join(input_dir, "*"))


class TestNS3D(unittest.TestCase):
    """Test fluiddyn.io.ns3d module."""

    def setUp(self):
        self._work_dir = "test_fluiddyn_io_ns3d"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        for path in input_files:
            copy(path, self._work_dir)

        os.chdir(self._work_dir)

        self.path_forcing_info_big = (
            "forcing_2D_info.in_L=30x30_nh=48_expLO_"
            "b=1.8_Ti=7.0_nbgene=4_d=4_T=28"
        )
        self.path_forcing_info_little = (
            self.path_forcing_info_big + "_little-endian"
        )
        self.path_field = "PV.t=0000.000"

    def tearDown(self):
        os.chdir("..")
        rmtree(self._work_dir)

    def test_forcing_info(self):
        f_b = NS3DForcingInfoFile(self.path_forcing_info_big)
        f_l = NS3DForcingInfoFile(self.path_forcing_info_little)

        self.assertEqual(f_l.byteorder, "little")
        self.assertEqual(f_b.byteorder, "big")

        with stdout_redirected():
            f_l.save_with_byteorder_changed()

    def test_field(self):
        f = NS3DFieldFile(self.path_field)
        f.read_xy()
        f.read_field()
        with stdout_redirected():
            f.save_with_byteorder_changed()
            f.save_with_resol_changed(2, 2, 2)


if __name__ == "__main__":
    unittest.main()



"""Utilities to build documentations with sphinx
================================================

The documentations of the packages of the Fluiddyn project are built and hosted
by Readthedocs (by the way, thank you to the readthedocs people!).

.. data:: on_rtd

   Boolean telling whether the code is being run on readthedocs server.

.. autofunction:: mock_modules

.. autosummary::
   :toctree:

   ipynb_maker
   fluidnbstripout
   fluiddocset
   mathmacro

"""

import os
import sys

try:
    from unittest.mock import Mock
except ImportError:
    # Python 2
    from mock import Mock


on_rtd = os.environ.get("READTHEDOCS")


class _MyMock(Mock):
    @classmethod
    def __getattr__(cls, name):
        return Mock()


def mock_modules(modules):
    """Mock modules (for example for building a documentation)

    Examples
    --------

    .. code-block:: python

       mock_modules((
           'h5py', 'h5netcdf', 'pyfftw', 'theano',
           'reikna.cluda', 'reikna.fft', 'reikna.transformations'))


    """

    sys.modules.update((mod_name, _MyMock()) for mod_name in modules)


__all__ = ["on_rtd", "mock_modules"]



"""fluidnbstripout utility
==========================

Very simple layer to stripout notebooks with the tool nbstripout
(https://github.com/kynan/nbstripout).

By default, fluidnbstripout excludes the notebooks whose name ends with
'.nbconvert.ipynb'.

"""

import argparse
import os
from glob import glob
import io

import nbstripout


def stripout(filename, keep_output=False, keep_count=False):
    print("stripout notebook", os.path.relpath(filename))
    try:
        with io.open(filename, "r", encoding="utf8") as f:
            nb = nbstripout.read(f, as_version=nbstripout.NO_CONVERT)
        nb = nbstripout.strip_output(nb, keep_output, keep_count)
        with io.open(filename, "w", encoding="utf8") as f:
            nbstripout.write(nb, f)
    except Exception:
        # Ignore exceptions for non-notebook files.
        print(f"Could not strip '{filename}'")
        raise


def main():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        "which",
        nargs="?",
        help="str indicating which notebooks have to be striped out.",
        type=str,
        default=os.getcwd(),
    )

    parser.add_argument(
        "-n",
        "--no_exclude_nbconvert",
        help='Do not exclude "nb_convert" notebooks.',
        action="store_true",
    )

    args = parser.parse_args()
    which = args.which

    if os.path.isfile(which):
        filename = which
        stripout(filename)
        return

    if os.path.isdir(which):
        which = os.path.join(which, "*.ipynb")

    filenames = glob(which)
    filenames.sort()

    if len(filenames) == 0:
        print(f'No notebooks found (input "{args.which}").')
        return

    # Big change compared to original nbstripout! :-)
    if not args.no_exclude_nbconvert:
        filenames = [
            fname for fname in filenames if not fname.endswith(".nbconvert.ipynb")
        ]

    for filename in filenames:
        stripout(filename)


if __name__ == "__main__":
    main()



"""
Utilities to query (:mod:`fluiddyn.io.query`)
=============================================

.. autofunction:: query_yes_no

"""

import builtins
import sys
import subprocess


def call_bash(commands):
    subprocess.call(["/bin/bash", "-c", commands])


def query_yes_no(question, default="yes"):
    """Ask a yes/no question and return the answer.

    Parameters
    ----------

    question : string
       String that is presented to the user.

    default : bool
       The default answer if the user just hits <Enter>.
       It must be "yes" (the default), "no" or None (meaning
       an answer is required of the user).

    Returns
    -------

    answer : bool
       The returned answer.
    """
    valid = {"yes": True, "y": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        sys.stdout.flush()
        choice = builtins.input().lower()
        if default is not None and choice == "":
            return valid[default]

        elif choice in valid:
            return valid[choice]

        else:
            sys.stdout.write(
                "Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n"
            )
            sys.stdout.flush()


def query(question, default=None):
    """Query an anwer to a general question."""
    sys.stdout.write(question)

    answer = builtins.input()

    if default is not None and answer == "":
        return default

    else:
        return answer


def num_from_str(s):
    """Return a num computed from a string."""
    try:
        return int(s)

    except ValueError:
        try:
            return float(s)

        except ValueError:
            raise ValueError("Can not convert the string into a numeric value.")


def query_number(question):
    """Query a number."""
    while True:
        sys.stdout.write(question + " ")
        answer = builtins.input()
        try:
            return num_from_str(answer)

        except ValueError:
            pass


def run_asking_agreement(command):
    """Use query_yes_no to ask if a command should be run."""

    question = 'Should the command "\n' + command + '\n" be run ?'
    if query_yes_no(question, default="no"):
        call_bash(command)



"""
Test dump utility
=================

"""

import unittest
import os
from shutil import rmtree
import sys

import numpy as np

import h5py
import h5netcdf

from ..redirect_stdout import stdout_redirected

from ..dump import main


class TestDump(unittest.TestCase):
    """Test fluiddyn.io.dump module."""

    _work_dir = "test_fluiddyn_io_dump"

    @classmethod
    def setUpClass(cls):
        if not os.path.exists(cls._work_dir):
            os.mkdir(cls._work_dir)

        os.chdir(cls._work_dir)

        # create nc file
        with h5netcdf.File("mydata.nc", "w") as f:
            nx = 5
            f.dimensions = {"x": nx}
            v = f.create_variable("hello", ("x",), float)
            v[:] = np.ones(nx)
            v = f.create_variable("/group0/data", ("y",), data=np.arange(10))
            v.attrs["foo"] = "bar"

        # create hdf5 file
        with h5py.File("mydata.h5", "w") as f:
            f.attrs["a"] = "a"
            f.create_dataset("a", [0, 1])
            g = f.create_group("mygroup")
            g.attrs["a"] = "a"

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test_nc(self):
        sys.argv = ["fluiddump", "mydata.nc"]
        with stdout_redirected():
            main()

    def test_h5(self):
        sys.argv = ["fluiddump", "mydata.h5"]
        with stdout_redirected():
            main()


if __name__ == "__main__":
    unittest.main()



"""
Test mycsv module
====================

"""

import unittest
import os
from shutil import rmtree

from ..mycsv import CSVFile


class TestDantex(unittest.TestCase):
    """Test fluiddyn.io.mycsv module."""

    def setUp(self):
        self._work_dir = "test_fluiddyn_io_mycsv"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        os.chdir(self._work_dir)

        with open("myfile.csv", "w") as f:
            f.write("a,b,c\n1,2,3\n")

    def tearDown(self):
        os.chdir("..")
        rmtree(self._work_dir)

    def test_csv(self):
        with CSVFile("myfile.csv") as f:
            f.get_fieldnames()
            f.load_as_dict(["a", "b"])


if __name__ == "__main__":
    unittest.main()



"""
Clusters
========

.. _clusters:
.. currentmodule:: fluiddyn.clusters

Provides:

.. autosummary::
   :toctree:

   oar
   slurm
   pbs
   local
   legi
   ciment
   cines
   snic

"""

import subprocess


class Cluster:
    _doc_commands = ""

    @classmethod
    def print_doc_commands(cls):
        print(cls._doc_commands)


def check_oar():
    """check if this script is run on a frontal with oar installed"""
    try:
        subprocess.check_call(["oarsub", "--version"], stdout=subprocess.PIPE)
        return True

    except OSError:
        return False


def check_pbs():
    """Check if this script is run on a frontal with pbs installed."""
    try:
        subprocess.check_call(["qsub", "--version"], stdout=subprocess.PIPE)
        return True

    except OSError:
        return False


def check_slurm():
    """Check if this script is run on a frontal with slurm installed."""
    try:
        subprocess.check_call(["sbatch", "--version"], stdout=subprocess.PIPE)
        return True

    except OSError:
        return False


help_docs = {
    "slurm": """sbatch
squeue -u $USER
scancel
scontrol hold <job_list>
scontrol release <job_list>
scontrol show job $JOBID
""",
    "pbs": """qsub
qstat -u $USER
qdel
qhold
qrls""",
    "oar": """oarsub -S script.sh
oarstat -u
oardel $JOB_ID
oarsub -C $JOB_ID""",
}


def print_help_scheduler(scheduler=None):
    """Detect the scheduler and print a minimal help."""
    if scheduler is None:
        if check_oar():
            scheduler = "oar"
        elif check_pbs():
            scheduler = "pbs"
        elif check_slurm():
            scheduler = "slurm"

    if scheduler:
        print(
            f"Scheduler detected: {scheduler}\n"
            + "Useful commands:\n----------------\n"
            + help_docs[scheduler]
        )
    else:
        print("No scheduler detected on this system.")



"""
Logger sending emails (:mod:`fluiddyn.util.logger`)
===================================================

Provides:

.. autoclass:: Logger
   :members:
   :private-members:

"""

from __future__ import print_function

from builtins import str
from builtins import object
import os
import sys
import time
import traceback

from fluiddyn.util import time_as_str
from fluiddyn.util.mail import send_email


class Logger:
    """Logger that can send emails."""

    def __init__(
        self,
        path="log.txt",
        email_to=None,
        email_from=None,
        email_title="title",
        email_delay=2 * 3600,
        email_server="localhost",
    ):

        if email_delay is None:
            email_delay = 2 * 3600

        self.path = path
        self.email_to = email_to
        self.has_to_send_email = email_to is not None

        self.email_from = email_from
        self.email_title = email_title
        self.email_delay = email_delay
        self.email_server = email_server
        self._normal_print = print
        self.time_last_email = time.time() - self.email_delay

        base, ext = os.path.splitext(self.path)
        self.path_logerr = base + "_stderr" + ext

        old_excepthook = sys.excepthook

        def my_excepthook(ex_cls, ex, tb):
            name_exception = ex_cls.__name__
            str_time = time_as_str()

            with open(self.path, "a") as f:
                f.write(
                    f"Python error ({name_exception}) at {str_time}\n"
                )

            with open(self.path_logerr, "a") as f:
                f.write(
                    "-" * 40
                    + f"\nError at {str_time}\n"
                    + "-" * 40
                    + "\n"
                    + "".join(traceback.format_exception(ex_cls, ex, tb))
                    + "\n"
                )

            if self.has_to_send_email and ex_cls != KeyboardInterrupt:
                self.send_email(name_exception=name_exception)

            old_excepthook(ex_cls, ex, tb)

        sys.excepthook = my_excepthook

    def print_log(self, *args, **kargs):
        """Replaces the Python 3 print function."""
        self._normal_print(*args, **kargs)
        self.write(*args, **kargs)

    def write(self, *args, **kargs):
        end = kargs.setdefault("end", "\n")
        with open(self.path, "a") as f:
            f.write(" ".join([str(arg) for arg in args]) + end)

    def send_email_if_has_to(self, name_exception=None, figures=None):
        """Sends an email if no email was sent recently."""
        if self.has_to_send_email:
            if time.time() - self.time_last_email >= self.email_delay:
                self.send_email(name_exception=name_exception, figures=figures)
                self.time_last_email = time.time()

    def send_email(self, name_exception=None, figures=None):
        """Sends the content of the storage file as an email"""
        with open(self.path, "r") as f:
            txt = f.read()

        if name_exception is None:
            subject = self.email_title
        else:
            subject = name_exception + ", " + self.email_title

            if os.path.exists(self.path_logerr):
                with open(self.path_logerr, "r") as f:
                    txt += f.read()

        if figures is None:
            files = None
        else:
            path_tmp = "/tmp/fluiddyn_" + time_as_str()
            os.makedirs(path_tmp)
            files = []
            for i, fig in enumerate(figures):
                fname = f"fig_{i}.png"
                path = os.path.join(path_tmp, fname)
                fig.savefig(path)
                files.append(path)

        send_email(
            subject,
            txt,
            address_recipients=self.email_to,
            address_sender=self.email_from,
            server=self.email_server,
            files=files,
        )

        print("Email sent at " + time_as_str())


# def __del__(self):
#     sys.excepthook = self._excepthook


if __name__ == "__main__":
    logger = Logger(
        "storage_file",
        "@ens-lyon.org",
        "@legi.cnrs.fr",
        email_title="itworks",
        email_delay=9,
    )
    print = logger.print_log
    print("fluiddyn, blablabla")

    logger.send_email_if_has_to()
    logger.send_email_if_has_to()
    time.sleep(10)
    logger.send_email_if_has_to()
    logger.send_email_if_has_to()



"""
Utilities for creating figures (:mod:`fluiddyn.output.figs`)
=============================================================

.. currentmodule:: fluiddyn.output.figs

Provides

.. autoclass:: Figure
   :members:

.. autoclass:: Figures
   :members:

"""

from __future__ import division, print_function

from builtins import str
from builtins import object
import sys
import os

import matplotlib.pyplot as plt
import matplotlib

from ..util import is_run_from_ipython
from .rcparams import set_rcparams


class Figures:
    """Represent a set of figures.

    Utilities to plot and save figures with matplotlib.

    Parameters
    ----------

    path_save : str
        Related to the path where to save.

    hastosave : bool
        If True, the function `Figure.saveifhasto` save the figure.

    for_beamer : bool
        If True, use beamer layout.

    for_article : bool
        If True, use article layout.

    fontsize : {18, int}
        Font size of the text in the figures.

    fontsize_pad : {9, int}
        Font size of the pad in the figures.
    """

    def __init__(
        self,
        path_save=None,
        hastosave=False,
        for_beamer=False,
        for_article=False,
        fontsize=18,
        fontsize_pad=9,
    ):

        self.hastosave = hastosave

        if path_save is None:
            self.path_save = os.getcwd()
        elif os.path.isabs(path_save):
            self.path_save = path_save
        else:
            self.path_save = os.path.join(os.getcwd(), path_save)

        set_rcparams(fontsize, for_article, for_beamer, fontsize_pad=9)

    def new_figure(
        self,
        name_file=None,
        num=None,
        fig_width_mm=200,
        fig_height_mm=150,
        size_axe=None,
    ):
        """Create a new Figure object and return it.

        Parameters
        ----------
        num : int, optional
            Number.

        fig_width_mm : {200, number}, optional
            Width (in mm)

        fig_height_mm : {150, number}, optional
            Height (in mm)

        size_axe : list, optional
            Size of the axe.

        name_file : str, optional
            Name of the file.

        """

        one_inch_in_mm = 25.4
        fig_width_inches = float(fig_width_mm) / one_inch_in_mm
        fig_height_inches = float(fig_height_mm) / one_inch_in_mm
        figsize = [fig_width_inches, fig_height_inches]

        dpi_latex = 72.27

        fig = Figure(
            num=num,
            figsize=figsize,
            dpi=dpi_latex,
            size_axe=size_axe,
            name_file=name_file,
            figures=self,
        )

        return fig


def show(block=None):
    """Show slightly more cleaver than old version of plt.show."""

    if block is None:
        if is_run_from_ipython():
            block = False
            plt.ion()
        else:
            block = True
            plt.ioff()

    if sys.platform.startswith("win") and not block:
        print("Warning: bug with anaconda and non-blocking show (?)")

    try:
        plt.show(block=block)
    except TypeError:
        plt.show()


class Figure(matplotlib.figure.Figure):
    """One figure.

    Improvement (?) of `matplotlib.figure.Figure`

    Parameters
    ----------
    (for the __init__ method)

    size_axe : list, optional
        Size of the axe.

    name_file : str, optional
        Name of the file.

    figures : :class:`fluiddyn.output.figs.Figures`
        Set of figures.

    kwargs : keyword arguments
        Given when create the figure.

    """

    def __init__(
        self, size_axe=None, name_file=None, figures=None, **kwargs  # *args,
    ):

        # Ugly workaround to be able to use the function plt.figure
        fig = plt.figure(**kwargs)
        for k, v in fig.__dict__.items():
            self.__dict__[k] = v

        if name_file is not None:
            self.name_file = name_file

        if figures is None:
            self.hastosave = True
            self.path_save = os.getcwd()
        else:
            self.hastosave = figures.hastosave
            self.path_save = figures.path_save
            self.figures = figures

        title = "Fig " + str(self.number)
        if name_file is not None:
            fig.name_file = name_file
            title += " " + name_file

        self.clf()
        self.canvas.set_window_title(title)

        if size_axe is not None:
            self.add_axes(size_axe)

    def saveifhasto(self, name_file=None, hastosave=None, verbose=True):
        """Save the figure if `hastosave` is True.

        Parameters
        ----------

        name_file : str, optional
            Name of the file (the extension has to indicate the format).

        hastosave : bool, optional
            If True, save the figure.

        verbose : {True, bool}, optional
             Print nothing if False.

        """

        if hastosave is None:
            hastosave = self.hastosave

        if hastosave:

            if name_file is None:
                try:
                    name_file = self.name_file
                except AttributeError:
                    raise ValueError("No name given...")

            if not os.path.exists(self.path_save):
                os.mkdir(self.path_save)

            path = os.path.join(self.path_save, name_file)

            if verbose:
                print("Save figure in file\n" + path)

            super().savefig(path)



import unittest

import numpy as np

from ..signal import decimate, FunctionLinInterp, deriv, smooth


class TestFFTW1DReal2Complex(unittest.TestCase):
    def test_signal(self):
        x = np.linspace(0, 2 * np.pi, 100)
        f = np.sin(x) + 0.02 * (np.random.rand(100) - 0.5)
        smooth(f, window_len=11, window="hanning")
        deriv(f, x, method="diff")
        deriv(f, x, dx=1, method="convolve")
        deriv(f, x, dx=1, method="gaussian_filter")

        sig = np.zeros([4, 4, 4])
        decimate(sig, 2, nwindow=2, axis=0)

        lin_interp = FunctionLinInterp(x, f)
        lin_interp(2)


if __name__ == "__main__":
    unittest.main()



"""Read and convert .im7 files (Davis Lavision) (:mod:`fluiddyn.io.davis`)
==========================================================================

"""

import os
from glob import glob
import argparse
from multiprocessing import Pool, cpu_count

try:
    import ReadIM

    _readim_ok = True
except ImportError as error:
    _error_readim = error
    _readim_ok = False


try:
    import png

    _png_ok = True
except ImportError as error:
    _error_png = error
    _png_ok = False


from .query import query_yes_no


def _import_error_readim():
    print(
        """ImportError ReadIM.
We need the package ReadIM to handle .im7 images. Please install it correctly
(see https://bitbucket.org/fleming79/readim)."""
    )
    raise ImportError(_error_readim)


def _import_error_png():
    print(
        """ImportError png.
We need the package pypng to handle png 'I;16'. Please install it correctly with
`pip install pypng` (see https://pypi.python.org/pypi/pypng)."""
    )
    raise ImportError(_error_png)


def readimages(path):

    if not _readim_ok:
        _import_error_readim()

    vbuff, vatts = ReadIM.extra.get_Buffer_andAttributeList(path)
    v_array, vbuff = ReadIM.extra.buffer_as_array(vbuff)

    images = []

    for iframe in range(vbuff.nf):
        images.append(v_array[iframe])

    return images


def parse_args():
    parser = argparse.ArgumentParser(
        description="""Convert .im7 files in a directory""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "path", nargs="?", default=os.getcwd(), help="Path of the directory."
    )

    parser.add_argument(
        "--sequential",
        help="Run sequencially, without multiprocessing.",
        action="store_true",
    )

    return parser.parse_args()


def convertim7(args):
    if not _readim_ok:
        _import_error_readim()

    if not _png_ok:
        _import_error_png()

    path = args.path

    if os.path.isdir(path):
        path = os.path.join(path, "*.im7")
    elif not path.endswith(".im7"):
        path += "*.im7"

    files = glob(path)
    files.sort()
    nb_files = len(files)

    if nb_files == 0:
        print("no file to treat.")
        return

    path_new_dir = os.path.dirname(files[0])
    path_new_dir += "_png"

    if nb_files == 1:
        plurial = ""
    else:
        plurial = "s"

    if not query_yes_no(
        "{} im7 file{} to be converted in the directory\n{}\n".format(
            nb_files, plurial, path_new_dir
        )
    ):
        return

    if not os.path.exists(path_new_dir):
        os.makedirs(path_new_dir)

    if args.sequential:
        for path in files:
            convert_1file(path, path_new_dir)
    else:
        pool = Pool(processes=min(cpu_count(), nb_files))
        pool.starmap(convert_1file, ((path, path_new_dir) for path in files))


def convert_1file(path, path_new_dir):

    name = os.path.split(path)[1]
    print("treat file " + name)
    name = os.path.splitext(name)[0]

    images = readimages(path)
    nb_images = len(images)

    if nb_images == 1:

        def get_new_name(i):
            return name + ".png"

    elif nb_images > 24:
        raise NotImplementedError

    else:
        isdecimal = name[-1].isdecimal()

        def get_new_name(i):
            if isdecimal:
                return name + chr(97 + i) + ".png"

            else:
                raise NotImplementedError

    for i, image in enumerate(images):
        new_name = get_new_name(i)
        print("save " + new_name)
        pathnew = os.path.join(path_new_dir, new_name)

        png_img = png.from_array(image, "L;16")
        png_img.save(pathnew)



"""
Constants (:mod:`fluiddyn.util.constants`)
==========================================


g = 9.81 m/s^2

nu_pure_water = 1.e-6 m^2/s

rho0 = 1. kg/m^3

rho_pure_water = 0.998 kg/m^3


"""

from __future__ import division, print_function

import numpy as np

g = 9.81  # (m/s^2)
nu_pure_water = 1.0e-6  # (m^2/s)
rho0 = 1.0  # (kg/m^3)
rho_pure_water = 0.998  # (kg/m^3)

del (division, print_function, np)



"""
Daemons (:mod:`fluiddyn.util.daemons`)
======================================

"""

from __future__ import division, print_function

from builtins import object
from threading import Thread
from multiprocessing import Process, Value

from fluiddyn.util import Params


class BaseDaemon:
    """Base Daemon class

    You may override the `run` method in a subclass and use
    `self.keepgoing.value`, `self._args` and `self._kwargs`.

    """

    def __init__(self, target=None, args=None, kwargs=None):
        self.daemon = True

    def stop(self):
        self.keepgoing.value = 0


class DaemonThread(BaseDaemon, Thread):
    def __init__(self, target=None, args=None, kwargs=None):
        Thread.__init__(self, target=target, args=args, kwargs=kwargs)
        super().__init__()
        # for compatibility with Process
        self.keepgoing = Params()
        self.keepgoing.value = True


class DaemonProcess(BaseDaemon, Process):
    def __init__(self, target=None, args=None, kwargs=None):
        Process.__init__(self, target=target, args=args, kwargs=kwargs)
        super().__init__()
        self.keepgoing = Value("i", 1)



"""Qt based GUI for parameters (:mod:`fluiddyn.util.paramcontainer_gui`)
========================================================================

Provides:

.. autoclass:: QtParamContainer
   :members:
   :private-members:

"""
from __future__ import division, print_function

import os
from copy import deepcopy
import subprocess

from matplotlib.backends.qt_compat import QtWidgets, QtCore

from fluiddyn.util import time_as_str

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:

    def _fromUtf8(s):
        return s


try:
    _encoding = QtWidgets.QApplication.UnicodeUTF8

    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(
            context, text, disambig, _encoding
        )


except AttributeError:

    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(context, text, disambig)


class QtParamContainer:
    """QWidget application framework for loading, editing, saving and launching
    a job from a ParamContainer object.

    """

    def __init__(
        self, params, top=False, module_run_from_xml="fluidimage.run_from_xml"
    ):

        self.params = deepcopy(params)
        self.module_run_from_xml = module_run_from_xml
        full_tag_dot = params._make_full_tag()
        self.full_tag = full_tag_dot.replace(".", "_")

        self.labels = {}
        self.lines_edit = {}
        self.buttons = {}

        self.qt_params_children = {}

        key_attribs = params._get_key_attribs()
        tag_children = params._tag_children

        self.page_main = QtWidgets.QWidget()
        self.page_main.setObjectName(_fromUtf8("page_main_" + self.full_tag))

        self.verticalLayout = QtWidgets.QVBoxLayout(self.page_main)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName(
            _fromUtf8("verticalLayout_" + self.full_tag)
        )

        if self.params._contains_doc():
            self.pushButton_doc = QtWidgets.QPushButton(self.page_main)
            self.pushButton_doc.setText("Display doc " + full_tag_dot)
            self.verticalLayout.addWidget(self.pushButton_doc)
            self.pushButton_doc.released.connect(self.params._print_docs)

        if len(key_attribs) > 0:
            self.page_attribs = QtWidgets.QWidget()
            self.page_attribs.setObjectName(
                _fromUtf8("page_attribs_" + self.full_tag)
            )

            self.formLayout_attribs = QtWidgets.QFormLayout(self.page_attribs)
            self.formLayout_attribs.setContentsMargins(0, 0, 0, 0)
            self.formLayout_attribs.setObjectName(
                _fromUtf8("formLayout_" + self.full_tag)
            )

            i = -1
            for key in key_attribs:
                i += 1
                label = self.labels[key] = QtWidgets.QLabel(self.page_attribs)
                label.setObjectName(
                    _fromUtf8("label_" + self.full_tag + "_" + key)
                )
                self.formLayout_attribs.setWidget(
                    i, QtWidgets.QFormLayout.LabelRole, label
                )

                line = self.lines_edit[key] = QtWidgets.QLineEdit(
                    self.page_attribs
                )
                line.setObjectName(_fromUtf8("line_" + self.full_tag + "_" + key))
                self.formLayout_attribs.setWidget(
                    i, QtWidgets.QFormLayout.FieldRole, line
                )

                label.setText(_translate("MainWindow", key, None))

                line.setText(
                    _translate("MainWindow", repr(self.params[key]), None)
                )

                if key == "path":
                    i += 1

                    def choose_name():
                        fileName = QtWidgets.QFileDialog.getOpenFileName(
                            self.page_attribs, "OpenFile"
                        )
                        self.lines_edit["path"].setText("'" + fileName + "'")

                    self.buttons[key] = QtWidgets.QPushButton(self.page_attribs)
                    self.formLayout_attribs.setWidget(
                        i, QtWidgets.QFormLayout.FieldRole, self.buttons[key]
                    )
                    self.buttons[key].setText("Navigate to choose the path")
                    self.buttons[key].released.connect(choose_name)

            self.verticalLayout.addWidget(self.page_attribs)

        if len(tag_children) > 0:
            self.toolBox = QtWidgets.QToolBox(self.page_main)

            for tag in tag_children:
                qtparam = self.qt_params_children[tag] = self.__class__(
                    self.params[tag]
                )

                self.toolBox.addItem(
                    qtparam.page_main, _fromUtf8(full_tag_dot + "." + tag)
                )

            self.verticalLayout.addWidget(self.toolBox)

        if top:
            self.pushButton_xml = QtWidgets.QPushButton(self.page_main)
            self.pushButton_xml.setText("Display as xml")
            self.verticalLayout.addWidget(self.pushButton_xml)
            self.pushButton_xml.released.connect(self.print_as_xml)

            self.pushButton_default = QtWidgets.QPushButton(self.page_main)
            self.pushButton_default.setText("Reset to default parameters")
            self.verticalLayout.addWidget(self.pushButton_default)
            self.pushButton_default.released.connect(self.reset_default_values)

            self.pushButton_launch = QtWidgets.QPushButton(self.page_main)
            self.pushButton_launch.setText("Launch computation")
            self.verticalLayout.addWidget(self.pushButton_launch)
            self.pushButton_launch.released.connect(self.launch)

    def set_values(self, params):
        for key in params._get_key_attribs():
            self.lines_edit[key].setText(
                _translate("MainWindow", repr(params[key]), None)
            )

        for tag in params._tag_children:
            self.qt_params_children[tag].set_values(params[tag])

    def reset_default_values(self):
        self.set_values(self.params)

    def produce_params(self):
        params = deepcopy(self.params)
        self.modif_params(params)
        return params

    def print_as_xml(self):
        params = self.produce_params()
        params._print_as_xml()

    def modif_params(self, params):
        for key in self.params._get_key_attribs():
            params[key] = eval(str(self.lines_edit[key].displayText()))

        for tag in self.params._tag_children:
            self.qt_params_children[tag].modif_params(params[tag])

    def launch(self):
        params = self.produce_params()

        path_dir = "tmp_fluiddyn_params"
        if not os.path.exists(path_dir):
            os.mkdir(path_dir)

        path = params._save_as_xml(
            path_file=os.path.join(
                path_dir, params._tag + time_as_str() + ".xml"
            ),
            find_new_name=True,
        )
        retcode = subprocess.call(
            ["python", "-m", self.module_run_from_xml, path]
        )
        return retcode



"""
IO for csv files (:mod:`fluiddyn.io.mycsv`)
===========================================

Provides:

.. autoclass:: CSVFile
   :members:
   :private-members:

"""

import csv

import numpy as np


class CSVFile:
    def __init__(self, *args, **kargs):
        self._textio = open(*args, **kargs)
        self.reader = csv.DictReader(self._textio)
        self.fieldnames = self.reader.fieldnames

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._textio.close()

    def load_as_dict(self, keys=None, skiptimes=0):
        if keys is None:
            keys = self.fieldnames
        else:
            for key in keys:
                if key not in self.fieldnames:
                    raise ValueError("A key is not in self.fieldnames")

        usecols = []
        for key in keys:
            usecols.append(self.fieldnames.index(key))

        # find dtypes from first line
        dtypes = []
        try:
            row = next(self.reader)
        except StopIteration:
            ret = {}
            for key in keys:
                ret[key] = np.array([])
            return ret

        else:
            for ik, key in enumerate(keys):
                value = row[key]
                if "." in value:
                    dtype = np.float64
                else:
                    dtype = np.int32
                dtypes.append(dtype)

        self._textio.seek(0)
        arr = np.loadtxt(
            self._textio, delimiter=",", skiprows=1 + skiptimes, usecols=usecols
        ).T

        ret = {}
        for ik, key in enumerate(keys):
            ret[key] = np.array(arr[ik], dtype=dtypes[ik])

        return ret

    def get_fieldnames(self):
        return self.fieldnames



import unittest

from ..colorchart import make_colorchart


class TestColorchart(unittest.TestCase):
    """Test fluiddyn.output.colorchart module."""

    def test_colorchart(self):
        make_colorchart(nb_colors=4, darkest_gray=0.5, lightest_gray=0.85)


if __name__ == "__main__":
    unittest.main()



"""
Test oar clusters
=================

"""
import unittest
import os
from shutil import rmtree
import subprocess

from ...io import stdout_redirected

from ..oar import ClusterOAR
from ..legi import Calcul, Calcul3, Calcul9, Calcul7, Calcul8
from ..ciment import Froggy

path_test = "tmp_test"

try:
    subprocess.check_call(["oarsub", "--version"], stdout=subprocess.PIPE)
    oar = True
except OSError:
    oar = False


class ClusterNoCheck(ClusterOAR):
    """A modified class which skips checking if oar is installed or not."""

    def check_oar(self):
        pass


class CalculNoCheck(ClusterNoCheck, Calcul):
    pass


class Calcul3NoCheck(ClusterNoCheck, Calcul3):
    pass


class Calcul9NoCheck(ClusterNoCheck, Calcul9):
    pass


class Calcul7NoCheck(ClusterNoCheck, Calcul7):
    pass


class Calcul8NoCheck(ClusterNoCheck, Calcul8):
    pass


class FroggyNoCheck(ClusterNoCheck, Froggy):
    pass


class TestCaseOAR(unittest.TestCase):
    Cluster = ClusterOAR
    ClusterNoCheck = ClusterNoCheck

    def setUp(self):
        self.cluster = self.Cluster()
        self.clusternocheck = self.ClusterNoCheck()

        self._work_dir = "tmp_test_oar"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)
        os.chdir(self._work_dir)

        with open("blabla.py", "w") as f:
            f.write('print("hello")')

    def tearDown(self):
        os.chdir("..")
        rmtree(self._work_dir)

    @unittest.skipIf(oar, "oar is present...")
    def test_submit_check(self):
        with self.assertRaises(OSError):
            self.cluster.submit_script("blabla.py", submit=False)

    def test_submit_nocheck(self):
        with self.assertRaises(ValueError):
            self.clusternocheck.submit_script(
                "script_that_does_not_exist.py", submit=False
            )

        with open("tmp_for_test.py", "w") as f:
            f.write('print("hello")')

        with stdout_redirected():
            self.clusternocheck.submit_script("tmp_for_test.py", submit=False)


class TestCaseCalcul(TestCaseOAR):
    Cluster = Calcul
    ClusterNoCheck = CalculNoCheck


class TestCaseCalcul3(TestCaseOAR):
    Cluster = Calcul3
    ClusterNoCheck = Calcul3NoCheck


class TestCaseCalcul9(TestCaseOAR):
    Cluster = Calcul9
    ClusterNoCheck = Calcul3NoCheck


class TestCaseCalcul7(TestCaseOAR):
    Cluster = Calcul7
    ClusterNoCheck = Calcul3NoCheck


class TestCaseCalcul8(TestCaseOAR):
    Cluster = Calcul8
    ClusterNoCheck = Calcul3NoCheck


class TestCaseFroggy(TestCaseOAR):
    Cluster = Froggy
    ClusterNoCheck = FroggyNoCheck


if __name__ == "__main__":
    unittest.main()



"""
Read and save image files
=========================


"""

from __future__ import print_function
import os

import numpy as np

try:
    from PIL import Image
except ImportError:
    pass

try:
    from cv2 import imread as _imread_opencv, IMREAD_ANYDEPTH

    use_opencv = True
except ImportError:
    use_opencv = False
    try:
        from imageio import imread as _imread
    except ImportError:
        # we use imread from matplotlib but there could be a problem with one
        # unittest.  If it is really a problem, install imageio.
        from matplotlib.pyplot import imread as _imread

try:
    from skimage.io import imread as _imread_ski
except ImportError:
    pass

try:
    import pims
except ImportError:
    pass

from .hdf5 import H5File


__all__ = ["imread", "imsave", "imread_h5", "imsave_h5", "extensions_movies"]


extensions_movies = ["cine", "im7"]


def imread(path, *args, **kwargs):
    """Wrapper for OpenCV/SciPy imread functions."""
    path = str(path)
    if path.endswith("]"):
        path, internal_index = path.split("[")
        internal_index = int(internal_index[:-1])
        for ext in extensions_movies:
            if path.endswith("." + ext):
                with pims.open(path) as images:
                    return images.get_frame(internal_index)

    if path.lower().endswith((".tiff", ".tif")):
        return _imread_ski(path, *args, **kwargs)

    if use_opencv:
        im = _imread_opencv(path, IMREAD_ANYDEPTH)
        if im is None:
            if not os.path.exists(path):
                raise FileNotFoundError(
                    f"No such file or directory: {path}"
                )
        return im
    else:
        return _imread(path, *args, **kwargs)


def _image_from_array(array, as_int):
    if as_int:
        if array.max() < 256:
            mode = "L"
            dtype = np.uint8
        else:
            mode = "I"
            dtype = np.uint32

        array = array.astype(dtype)
    else:
        dtype = array.dtype
        if np.issubdtype(dtype, np.floating):
            mode = "F"
        elif np.issubdtype(dtype, np.uint8):
            mode = "L"
        elif np.issubdtype(dtype, np.integer):
            mode = "I"
        else:
            raise NotImplementedError("Unexpected dtype %s" % dtype)

    # im = toimage(arr=array, mode=mode, channel_axis=2)
    # print('in _image_from_array', mode, array)

    im = Image.fromarray(array, mode)
    return im


def imsave(path, array, format=None, as_int=False):
    """
    Alternative implementation of `scipy.misc.imsave` function.
    Detects a compatible format based on the array dtype, rather than relying
    on the file extension.

    .. WARNING: setting `as_int=True` might lead to loss of precision.

    """

    im = _image_from_array(array, as_int)

    path = str(path)

    if format is None:
        if im.mode == "F" or any(
            [path.endswith(ext) for ext in (".tif", ".tiff")]
        ):
            format = "TIFF"
        else:
            format = "PNG"

    if format.lower() == "tiff":
        if any(
            [path.endswith(ext) for ext in (".png", ".PNG")]
        ) and np.issubdtype(array.dtype, np.floating):
            print(
                "warning: can not save float image as png. " "Using tif format."
            )

        if not any([path.endswith(ext) for ext in (".tif", ".tiff")]):
            path += ".tiff"
    elif format.lower == "png":
        if not any([path.endswith(ext) for ext in (".png", ".PNG")]):
            if path.endswith(".tif"):
                path = path[: -len(".tif")]
            if path.endswith(".tiff"):
                path = path[: -len(".tiff")]
            path += ".png"

    im.save(path, format)
    im.close()


def imread_h5(path):
    """Read image(s) stored in a HDF5 file."""

    with ImageH5File(path, "r") as f:
        return f.load(group="images")


def imsave_h5(
    path,
    array,
    params=None,
    attrs={},
    compression="gzip",
    as_int=False,
    splitext=True,
):
    """Saves an image as a compressed HDF5 file."""

    fname = os.path.basename(path)
    if splitext:
        root, ext = os.path.splitext(path)
    else:
        root = path
    h5path = str(root) + ".h5"

    if as_int:
        if array.max() < 256:
            dtype = np.uint8
        else:
            dtype = np.uint32

        array = array.astype(dtype)

    with ImageH5File(h5path, "w") as f:
        f.save_dict(
            "images", {fname: array}, compression=compression, shuffle=True
        )
        f.save_attrs(attrs)
        if params is not None:
            params._save_as_hdf5(hdf5_parent=f)


class ImageH5File(H5File):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load(self, group):
        """Load images in the HDF5 file."""

        images = self[group]
        nb_images = len(images)
        if nb_images == 1:
            dset = list(images.items())[0][1]
            return dset[...]

        else:
            dico = {}
            for k, v in images.items():
                dico[k] = v[...]

            return dico

    def save_attrs(self, dictattrs):
        """Save the `dictattrs` as attributes."""

        for k, v in list(dictattrs.items()):
            self.attrs[k] = v

    def save_dict(self, keydict, dicttosave, **kwargs):
        """Save the dictionary `dicttosave` in the file."""

        group = self.create_group(keydict)

        if len(dicttosave) > 0:
            for k, v in list(dicttosave.items()):
                group.create_dataset(k, data=v, **kwargs)



"""Terminal color codes (:mod:`fluiddyn.util.terminal_colors`)
==============================================================

Defines string variables useful to print in color in a terminal.

"""

from __future__ import print_function

HEADER = "\033[95m"
OKBLUE = "\033[94m"
OKGREEN = "\033[92m"
WARNING = "\033[93m"
FAIL = "\033[91m"

BLACK = "\033[30m"
RED = "\033[31m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
BLUE = "\033[34m"
MAGENTA = "\033[35m"
CYAN = "\033[36m"
WHITE = "\033[37m"

ENDC = "\033[0m"
BOLD = "\033[1m"
UNDERLINE = "\033[4m"

_color_dict = vars()


def print_fail(*args, **kwargs):
    """Print in bold with "FAIL" font color."""
    args, kwargs = _colorize(color="FAIL", bold=True, *args, **kwargs)
    print(*args, **kwargs)


def print_warning(*args, **kwargs):
    """Print with "WARNING" font colour."""
    args, kwargs = _colorize(color="WARNING", *args, **kwargs)
    print(*args, **kwargs)


def _colorize(*args, **kwargs):
    """Wrap the args with a specified terminal font color."""
    try:
        color = kwargs.pop("color")
        try:
            bold = kwargs.pop("bold")
        except KeyError:
            bold = False

        color = _color_dict[color]

        if len(args) > 0:
            args = list(args)
            if bold:
                color += BOLD
            args[0] = "{}{}".format(color, args[0])
            args[-1] = "{}{}".format(args[-1], ENDC)

    finally:
        return args, kwargs


def cstring(*args, **kwargs):
    """Return a colored string.

    Parameters
    ----------
    args : iterable of str
    bold : bool
    color : str, {'HEADER', 'OKBLUE', 'OKGREEN', 'WARNING', 'FAIL', 'ENDC',
                  'BOLD', 'UNDERLINE' 'BLACK', 'RED', 'GREEN', 'YELLOW', 'BLUE',
                  'MAGENTA', 'CYAN', 'WHITE'}
        Terminal color to use.

    """
    args, kwargs = _colorize(*args, **kwargs)
    cstr = " ".join(args)
    return cstr


def cprint(*args, **kwargs):
    """Print with a specified terminal font color.

    Parameters
    ----------
    args : iterable
        To be passed into print_function.
    bold : bool
    color : str, {'HEADER', 'OKBLUE', 'OKGREEN', 'WARNING', 'FAIL', 'ENDC',
                  'BOLD', 'UNDERLINE' 'BLACK', 'RED', 'GREEN', 'YELLOW', 'BLUE',
                  'MAGENTA', 'CYAN', 'WHITE'}

        Terminal color to use.

    """
    args, kwargs = _colorize(*args, **kwargs)
    print(*args, **kwargs)



"""
IO for Digiflow files (:mod:`fluiddyn.io.digiflow`)
===================================================

.. currentmodule:: fluiddyn.io.digiflow

Provides the classes :class:`DigiflowImage` and :class:`DigiflowMovie`.

.. autoclass:: DigiflowImage
   :members:

.. autoclass:: DigiflowMovie
   :members:

"""

from __future__ import division, print_function

from builtins import range
from builtins import object
import numpy as np
import matplotlib.pyplot as plt

from fluiddyn.io.binary import BinFile
from fluiddyn.util import Params

ddatatypes = {
    0x1001: "8 bit image",
    0x11001: "8 bit multi-plane image",
    0x12001: "Compressed 8 bit image",
    0x1004: "32 bit image",
    0x11004: "32 bit multi-plane image",
    0x12004: "Compressed 32 bit image",
    0x1008: "64 bit image",
    0x11008: "64 bit multi-plane image",
    0x12008: "Compressed 64 bit image",
    0x1014: "32 bit range",
    0x1018: "64 bit range",
    0x1100: "Rescale image",
    0x1101: "Rescale image rectangle",
    0x2000: "Colour scheme",
    0x2001: "Colour scheme name",
    0x2003: "Colour scheme name variable",
    0x3000: "Description",
    0x3001: "User comments",
    0x3002: "Creating process",
    0x3003: "Creator details",
    0x3018: "Image time",
    0x4008: "Image coordinates",
    0x4108: "Image plane details",
}

keys_image = [
    "8 bit image",
    "8 bit multi-plane image",
    "Compressed 8 bit image",
    "32 bit image",
    "32 bit multi-plane image",
    "Compressed 32 bit image",
    "64 bit image",
    "64 bit multi-plane image",
    "Compressed 64 bit image",
]


class DigiflowImage:
    """A digiflow image (.dfi, "image" containing 3 scalar fields).


    """

    def __init__(self, path_file=None):

        if path_file is not None:
            self.path_file = path_file
            self._load()

    def _load(self):
        """Loads the file."""

        with BinFile(self.path_file) as f:
            # header of the file:
            id_format = f.readt(32, "s")
            version = f.readt(1, "I")

            if id_format != "Tagged floating point image file":
                raise ValueError(
                    "This file does not seem to be a valid dfi file."
                )

            self.dict_infos = {}
            self._datatype_codes = []
            while self._read_one_field(f):
                pass

    def _read_one_field(self, f):
        """Read one field in the file `f`."""
        # tag for one field
        datatype_code = f.readt(1, "uint32")
        if datatype_code == "eof":
            # All the file has been read.
            return False

        self._datatype_codes.append(datatype_code)
        nbytes = f.readt(1, "uint32")

        datatype = ddatatypes[datatype_code]
        # print(datatype)

        if datatype in keys_image:
            self._read_one_image(datatype, nbytes, f)

        elif datatype == "32 bit range":
            assert nbytes == 8
            r_black, r_white = f.readt(2, "float32")
            self.range_data = np.array([r_black, r_white])

        elif datatype == "64 bit range":
            assert nbytes == 16
            r_black, r_white = f.readt(2, "float64")
            self.range_data = np.array([r_black, r_white])

        elif datatype == "Rescale image":
            print(datatype, "not yet written")

        elif datatype == "Rescale image rectangle":
            assert nbytes == 32
            self.rectangle_rescaleim = Params()
            r = self.rectangle_rescaleim
            (
                r.nxwant,
                r.nywant,
                r.method,
                r.userectangle,
                r.xmin,
                r.ymin,
                r.xmax,
                r.ymax,
            ) = f.readt(8, "uint32")

        elif datatype == "Colour scheme":
            assert nbytes == 3 * 256 * 1
            self.colours = np.array(f.readt(3 * 256, "uint8")).reshape([3, 256])

        elif datatype == "Colour scheme name":
            print(datatype, "not yet written")

        elif datatype == "Colour scheme name variable":
            print(datatype, "not yet written")

        elif datatype == "Description":
            assert nbytes == 512
            self.dict_infos[datatype] = f.readt(512, "s")

        elif datatype in ["User comments", "Creating process"]:
            nb_bytes_descrition = f.readt(1, "uint32")
            assert nbytes == nb_bytes_descrition + 4
            self.dict_infos[datatype] = f.readt(nb_bytes_descrition, "s")

        elif datatype == "Creator details":
            assert nbytes == 64 + 48 + 32 * 5 + 16 * 2
            self.dict_infos["DigiFlow"] = f.readt(32, "s")
            self.dict_infos["buildDate"] = f.readt(16, "s")
            self.dict_infos["licenceType"] = f.readt(16, "s")
            self.dict_infos["nameUser"] = f.readt(32, "s")
            self.dict_infos["nameComputer"] = f.readt(32, "s")
            self.dict_infos["nameDomain"] = f.readt(32, "s")
            self.dict_infos["guidUser"] = f.readt(32, "s")
            self.dict_infos["macAddress"] = f.readt(48, "s")
            self.dict_infos["ipAddress"] = f.readt(64, "s")

        elif datatype == "Image time":
            assert nbytes == 32
            self.times = Params()
            t = self.times
            t.iframe, t.reserved = f.readt(2, "uint32")
            t.time, t.tstep, t.tfirst = f.readt(3, "float64")

        elif datatype == "Image coordinates":
            assert nbytes == 132
            self.coordinates = Params()
            c = self.coordinates
            c.kind = f.readt(1, "uint32")
            xworldperpixel, yworldperpixel, xoriginworld, yoriginworld = f.readt(
                4, "float64"
            )
            c.worldsperpixel = np.array([xworldperpixel, yworldperpixel])
            c.origins_world = np.array([xoriginworld, yoriginworld])
            c.xunits = f.readt(16, "s")
            c.yunits = f.readt(16, "s")
            c.original_name = f.readt(64, "s")

        elif datatype == "Image plane details":
            nplanes = f.readt(1, "uint32")
            assert nbytes == 100 * nplanes + 4
            self.infos_planes = Params()
            infop = self.infos_planes
            infop.codes = []
            infop.keys = []
            infop.params = []
            infop.dunno = []
            for ip in range(nplanes):
                infop.codes.append(int("{:x}".format(f.readt(1, "uint32"))))
                infop.keys.append(f.readt(32, "s"))
                infop.params.append(list(f.readt(4, "float64")))
                infop.dunno.append(f.readt(32, "s"))

        return True

    def _read_one_image(self, datatype, nbytes, f):
        """Reads one image."""

        nx, ny = f.readt(2, "uint32")
        if datatype == "8 bit image":
            assert nbytes == nx * ny + 8
            self.data = np.array(f.readt(nx * ny, "uint8")).reshape([1, ny, nx])

        elif datatype == "8 bit multi-plane image":
            nz = f.readt(1, "uint32")
            assert nbytes == nx * ny * nz + 12
            self.data = np.array(f.readt(nx * ny * nz, "uint8")).reshape(
                [nz, ny, nx]
            )

        elif datatype == "Compressed 8 bit image":
            nz, size_compressed = f.readt(2, "uint32")
            assert nbytes == size_compressed + 16
            self.data = np.array(
                f.readt_zlib(size_compressed, nx * ny * nz, "uint8")
            ).reshape([nz, ny, nx])

        elif datatype == "32 bit image":
            assert nbytes == 4 * nx * ny + 8
            self.data = np.array(f.readt(nx * ny, "float32")).reshape([1, ny, nx])

        elif datatype == "32 bit multi-plane image":
            nz = f.readt(1, "uint32")
            assert nbytes == 4 * nx * ny * nz + 12
            self.data = np.array(f.readt(nx * ny * nz, "float32")).reshape(
                [nz, ny, nx]
            )

        elif datatype == "Compressed 32 bit image":
            nz, size_compressed = f.readt(2, "uint32")
            assert nbytes == size_compressed + 16
            self.data = np.array(
                f.readt_zlib(size_compressed, nx * ny * nz, "float32")
            ).reshape([nz, ny, nx])

        elif datatype == "64 bit image":
            assert nbytes == 8 * nx * ny + 8
            self.data = np.array(f.readt(nx * ny, "float64")).reshape([1, ny, nx])

        elif datatype == "64 bit multi-plane image":
            nz = f.readt(1, "uint32")
            assert nbytes == 8 * nx * ny * nz + 12
            self.data = np.array(f.readt(nx * ny * nz, "float64")).reshape(
                [nz, ny, nx]
            )

        elif datatype == "Compressed 64 bit image":
            nz, size_compressed = f.readt(2, "uint32")
            assert nbytes == size_compressed + 16
            self.data = np.array(
                f.readt_zlib(size_compressed, nx * ny * nz, "float64")
            ).reshape([nz, ny, nx])

    # def save(self):
    #     print('Not yet implemented...')

    def __getitem__(self, key):
        if key not in self.infos_planes.keys:
            raise KeyError("correct keys:", self.infos_planes.keys)

        return self.data[self.infos_planes.keys.index(key)]


class DigiflowMovie:
    """A digiflow movie (.dfm, "movies", set of images).

    """

    def __init__(self, path_file=None):
        if path_file is not None:
            self.path_file = path_file
            self._load_info()

            self.deltat = self.movie_header["dtSampleSpacing"]

    def _load_info(self):
        """Loads the file."""

        with BinFile(self.path_file) as f:
            # do not change the order of these lines!
            # do not comment one of these lines!
            self._read_file_header(f)
            self._read_hist_info(f)
            self._read_movie_header(f)

            nMovieFrames = self.movie_header["nMovieFrames"]
            temp = f.readt(2 * nMovieFrames, "int64")
            temp = np.array(temp).reshape([nMovieFrames, 2])
            self.iFrameNumber = temp[:, 0]
            self.iPtrFrame = temp[:, 1]

            self.nb_frames = len(self.iFrameNumber)

    def _read_file_header(self, f):
        # header of the file:
        fileowner = f.readt(8, "s")
        version = f.readt(8, "s")
        iPtrHistory = f.readt(1, "I")
        filetype = f.readt(16, "s")
        comments = f.readt(220, "s")
        d = locals()
        del (d["f"], d["self"])
        self.file_header = d

    def _read_hist_info(self, f):
        # History Header Information
        iPtrPrivateHeader = f.readt(1, "I")
        iDummy = f.readt(1, "I")
        CreatedBy = f.readt(8, "s")
        Hversion = f.readt(8, "s")
        CreatedUser = f.readt(16, "s")
        CreatedName = f.readt(64, "s")
        CreatedDate = f.readt(8, "s")
        CreatedTime = f.readt(8, "s")
        ModifiedUser = f.readt(16, "s")
        ModifiedName = f.readt(64, "s")
        ModifiedDate = f.readt(8, "s")
        ModifiedTime = f.readt(8, "s")
        UnUsed = f.readt(40, "s")
        del (UnUsed)
        d = locals()
        del (d["f"], d["self"])
        self.hist_info = d

    def _read_movie_header(self, f):
        # Movie Header Information

        (iFormatType, iFrameRate) = f.readt(2, "uint16")
        (iSampleSpacing, iMovieDuration, iPtrFrameTable, nMovieFrames) = f.readt(
            4, "uint32"
        )
        (iw0, iw1, jw0, jw1, idi, jdj) = f.readt(6, "uint16")

        self.shape_im = [iw1 - iw0 + 1, jw1 - jw0 + 1]
        self.size_im = self.shape_im[0] * self.shape_im[1]

        nSize = f.readt(1, "uint32")
        AspectRatio = f.readt(1, "float32")

        nBits = f.readt(1, "uint16")

        iOLUTRed = np.array(f.readt(256, "uint8"))
        iOLUTGreen = np.array(f.readt(256, "uint8"))
        iOLUTBlue = np.array(f.readt(256, "uint8"))
        nFrameTableLength = f.readt(1, "uint32")
        RecordAtFieldSpacing = f.readt(1, "uint16")

        dtSampleSpacing = f.readt(1, "float32")
        UnUsed = f.readt(204, "uint8")
        del (UnUsed)

        d = locals()
        del (d["f"], d["self"])
        self.movie_header = d

    def __getitem__(self, arg):

        # itstart = 0
        # itstop = self.nb_frames-1
        # itstep = 1

        # iystart = 0
        # iystop = self.shape_im[0]
        # iystep = 1

        # ixstart = 0
        # ixstop = self.shape_im[1]
        # ixstep = 1

        if isinstance(arg, int):
            with open(self.path_file) as f:
                f.seek(self.iPtrFrame[arg])
                return np.fromfile(f, dtype=np.uint8, count=self.size_im).reshape(
                    self.shape_im
                )

        elif isinstance(arg, slice):
            raise NotImplementedError

        # itstart = arg.start
        # itstop = arg.stop
        # itstep = arg.step

        elif isinstance(arg, tuple):
            raise NotImplementedError

    # itstart = arg[0].start
    # itstop = arg[0].stop
    # itstep = arg[0].step

    # nb_values_between_frames = self.iPtrFrame[1] - self.iPtrFrame[0]

    # shape_ret = [nb_frames] + list(self.shape_im)
    # size_im = self.size_im
    # nb_values_to_jump_startim = 0
    # nb_values_to_jump_endim = nb_values_between_frames - size_im

    # ret = np.empty(shape_ret)

    # with BinFile(self.path_file) as f:
    #     f.seek(self.iPtrFrame[iframe_start])
    #     for iframe in xrange(nb_frames):
    #         # temp = np.array(f.readt(self.size_im, 'uint8'))
    #         # ret[iframe].flat = temp#.reshape(self.shape_im)
    #         f.seek(nb_values_to_jump_startim, 1)
    #         ret[iframe].flat = f.readt(size_im, 'uint8')
    #         f.seek(nb_values_to_jump_endim, 1)

    # if nb_frames == 1:
    #     return ret[0]
    # else:
    #     return ret

    # def load_contiguous_frames2(self, iframe_start, nb_frames=1):

    #     if iframe_start+1 > self.nb_frames:
    #         raise ValueError('A non excisting frame has been asked.')

    #     if iframe_start+nb_frames > self.nb_frames:
    #         print('too many frames asked...')
    #         nb_frames = self.nb_frames - iframe_start

    #     ret = np.empty([nb_frames]+list(self.shape_im))

    #     if self.nb_frames > 1:
    #         nb_values_between_frames = self.iPtrFrame[1] - self.iPtrFrame[0]
    #     else:
    #         nb_values_between_frames = self.size_im

    #     with BinFile(self.path_file) as f:
    #         f.seek(self.iPtrFrame[iframe_start])
    #         temp = f.readt(nb_frames*nb_values_between_frames, 'uint8')

    #     for iframe in xrange(nb_frames):
    #         istart = iframe*nb_values_between_frames
    #         temp2 = np.array(temp[istart:istart+self.size_im])
    #         ret[iframe] = temp2.reshape(self.shape_im)

    #     if nb_frames == 1:
    #         return ret[0]
    #     else:
    #         return ret

    def load_contiguous_frames(self, iframe_start, nb_frames=1):

        if iframe_start + 1 > self.nb_frames:
            raise ValueError("A non-existing frame has been asked.")

        if iframe_start + nb_frames > self.nb_frames:
            print("too many frames asked...")
            nb_frames = self.nb_frames - iframe_start

        ret = np.empty([nb_frames] + list(self.shape_im))

        if self.nb_frames > 1:
            nb_values_to_jump = (
                self.iPtrFrame[1] - self.iPtrFrame[0] - self.size_im
            )
        else:
            nb_values_to_jump = 0

        # with BinFile(self.path_file) as f:
        #     f.seek(self.iPtrFrame[iframe_start])
        #     for iframe in xrange(nb_frames):
        #         # ret[iframe].flat = f.readt(self.size_im, 'uint8')
        #         ret[iframe].flat = np.fromfile(
        #             f, dtype=np.uint8, count=self.size_im)
        #         f.seek(nb_values_to_jump, 1)

        with open(self.path_file, "rb") as f:
            f.seek(self.iPtrFrame[iframe_start])
            for iframe in range(nb_frames):
                # ret[iframe].flat = f.readt(self.size_im, 'uint8')
                ret[iframe].flat = np.fromfile(
                    f, dtype=np.uint8, count=self.size_im
                )
                f.seek(nb_values_to_jump, 1)

        if nb_frames == 1:
            return ret[0]

        else:
            return ret

    def _test_end_of_file(self, i):
        "Seems to reach eof if i > 275 but why ?"
        with BinFile(self.path_file) as f:
            f.seek(self.iPtrFrame[-1])
            im = f.readt(self.size_im, "uint8")

            print("last values in last image:\n", im[-10:])
            print("after the last values:\n", f.readt(i, "uint8"))

    def show_movie(self, i_start, i_stop, i_step=1, fps=4, decimate=4):
        plt.ion()
        fig = plt.figure()
        ax = plt.gca()

        data = self[i_start]
        if decimate > 1:
            data = data[::decimate, ::decimate]

        quadmesh = ax.pcolormesh(data)
        quadmesh.set_clim([0.0, 255.0])

        ax.set_xlim([0, data.shape[1]])
        ax.set_ylim([0, data.shape[0]])

        for i in range(i_start + i_step, i_stop, i_step):
            with open(self.path_file) as f:
                f.seek(self.iPtrFrame[i])
                data = np.fromfile(f, dtype=np.uint8, count=self.size_im).reshape(
                    self.shape_im
                )

                if decimate > 1:
                    data = data[::decimate, ::decimate]

            quadmesh.set_array(data.ravel())
            fig.canvas.draw()

    def plot_image(self, i_image, decimate=1, clim=None, cmap=None):
        plt.ion()
        plt.figure()
        ax = plt.gca()

        data = self[i_image]
        if decimate > 1:
            data = data[::decimate, ::decimate]

        # if cmap is None:
        #     cmap = plt.cm.hsv

        quadmesh = ax.pcolormesh(data, cmap=cmap, shading="flat")

        if clim is not None:
            quadmesh.set_clim(clim)

        ax.set_xlim([0, data.shape[1]])
        ax.set_ylim([0, data.shape[0]])

        plt.show()

    def make_time_serie(
        self, i_start, i_stop, i_x, i_step=1, has_to_plot=True, cmap=None
    ):

        if i_stop > self.nb_frames - 1:
            i_stop = self.nb_frames - 1

        inds_t = list(range(i_start, i_stop + 1, i_step))

        time_serie = np.zeros([self.shape_im[0], len(inds_t)])

        for it in inds_t:
            a = self[it]
            time_serie[:, it] = a[:, i_x]

        if has_to_plot:
            plt.ion()
            plt.figure()
            ax = plt.gca()
            ax.pcolormesh(time_serie, cmap=cmap, shading="flat")
            plt.show()

        return time_serie


if __name__ == "__main__":

    def plot_im(im):
        import fluiddyn.figs as figs

        import matplotlib.pyplot as plt

        figures = figs.Figures(hastosave=False, for_beamer=False, fontsize=20)
        size_axe = [0.15, 0.12, 0.8, 0.76]

        fig = figures.new_figure(
            name_file="fig_im",
            fig_width_mm=170,
            fig_height_mm=220,
            size_axe=size_axe,
        )

        ax1 = fig.gca()

        ax1.set_xlabel(r"$x$ (pixels)")
        ax1.set_ylabel(r"$z$ (pixels)")

        pc = ax1.pcolormesh(im, cmap=plt.cm.hsv, shading="flat")

        fig.colorbar(pc)

        ax1.set_xlim([0, im.shape[1]])
        ax1.set_ylim([0, im.shape[0]])

        figs.show()

    path_file = (
        "/home/pa371/Data_reasons_pa371/Dropbox"
        "/Strat_inclined_duct/Decomposition/Data_exp/Exp_m7"
        "/clean_m7_lif3_0036.dfi"
    )

    print(path_file)
    dfi = DigiflowImage(path_file)

    plot_im(dfi.data[0])

# dfm = DigiflowMovie(path_file)
# im = dfm.load_contiguous_frames(10, nb_frames=3)
